# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

import datetime

import numpy as np
from typing import Dict, List, Union
from PySide6 import QtCore, QtWidgets, QtGui
from enum import EnumMeta
from VeraGrid.Gui.gui_functions import (IntDelegate, ComboDelegate, TextDelegate, FloatDelegate, ColorPickerDelegate,
                                        ComplexDelegate, LineLocationsDelegate, DateTimeDelegate)
from VeraGrid.Gui.wrappable_table_model import WrappableTableModel
from VeraGridEngine.Devices import Bus, ContingencyGroup
from VeraGridEngine.Devices.Parents.editable_device import GCProp, GCPROP_TYPES
from VeraGridEngine.enumerations import DeviceType
from VeraGridEngine.Devices.Branches.line_locations import LineLocations
from VeraGridEngine.Devices.types import ALL_DEV_TYPES


class ObjectsModel(WrappableTableModel):
    """
    Class to populate a Qt table view with the properties of objects
    """

    def __init__(self,
                 objects: List[ALL_DEV_TYPES],
                 property_list: List[GCProp],
                 time_index: Union[int, None],
                 parent: QtWidgets.QTableView = None,
                 editable=False,
                 transposed=False,
                 check_unique: Union[None, List[str]] = None,
                 dictionary_of_lists: Union[None, Dict[DeviceType, List[ALL_DEV_TYPES]]] = None):
        """

        :param objects: list of objects associated to the editor
        :param property_list: List with the declared properties of an object
        :param parent: Parent object: the QTableView object
        :param editable: Is the table editable?
        :param transposed: Display the table transposed?
        :param dictionary_of_lists: dictionary of lists for the Delegates
        """
        WrappableTableModel.__init__(self, parent)

        self.parent = parent

        self.time_index_: Union[int, None] = time_index

        self.editable = editable

        self.objects: List[ALL_DEV_TYPES] = objects

        self.property_list: List[GCProp] = list()
        self.attributes: List[str] = list()
        self.attribute_types: List[GCPROP_TYPES] = list()
        self.units: List[str] = list()
        self.tips: List[str] = list()
        self.non_editable_attributes: List[str] = list()

        for p in property_list:
            if p.display:
                self.property_list.append(p)
                self.attributes.append(p.name)
                self.attribute_types.append(p.tpe)
                self.units.append(p.units)
                self.tips.append(p.definition)

                if not p.editable:
                    self.non_editable_attributes.append(p.name)

        self.check_unique = check_unique if check_unique is not None else list()

        self.r = len(self.objects)

        self.c = len(self.attributes)

        self.formatter = lambda x: "%.2f" % x

        self.transposed = transposed

        self.dictionary_of_lists = dictionary_of_lists if dictionary_of_lists is not None else dict()

        self.set_delegates()

    def set_time_index(self, time_index: Union[int, None]):
        """
        Set the time index of the table
        :param time_index: None or integer value
        """
        self.time_index_ = time_index
        role = 0
        index = QtCore.QModelIndex()
        self.dataChanged.emit(index, index, [role])

    def set_delegates(self) -> None:
        """
        Set the cell editor types depending on the attribute_types array
        """

        if self.transposed:
            F = self.parent.setItemDelegateForRow
        else:
            F = self.parent.setItemDelegateForColumn

        for i in range(self.c):

            if self.property_list[i].is_color:
                delegate = ColorPickerDelegate(self.parent)
                F(i, delegate)

            elif self.property_list[i].is_date:
                delegate = DateTimeDelegate(self.parent)
                F(i, delegate)

            else:
                tpe = self.attribute_types[i]

                if tpe is bool:
                    delegate = ComboDelegate(self.parent, [True, False], ['True', 'False'])
                    F(i, delegate)

                elif tpe is str:
                    delegate = TextDelegate(self.parent)
                    F(i, delegate)

                elif tpe is float:
                    delegate = FloatDelegate(self.parent)
                    F(i, delegate)

                elif tpe is int:
                    delegate = IntDelegate(self.parent)
                    F(i, delegate)

                elif tpe is complex:
                    delegate = ComplexDelegate(self.parent)
                    F(i, delegate)

                elif tpe is LineLocations:
                    delegate = LineLocationsDelegate(self.parent)
                    F(i, delegate)

                elif tpe is None:
                    F(i, None)
                    if len(self.non_editable_attributes) == 0:
                        self.non_editable_attributes.append(self.attributes[i])

                elif isinstance(tpe, EnumMeta):
                    objects = list(tpe)
                    values = [x.value for x in objects]
                    delegate = ComboDelegate(self.parent, objects, values)
                    F(i, delegate)

                elif tpe in self.dictionary_of_lists:
                    # foreign key objects drop-down
                    objs = self.dictionary_of_lists[tpe]
                    delegate = ComboDelegate(parent=self.parent,
                                             objects=[None] + objs,
                                             object_names=['None'] + [x.name for x in objs])
                    F(i, delegate)

                else:
                    F(i, None)

    def update(self):
        """
        update table
        """
        self.r = len(self.objects)
        # row = self.rowCount()
        # self.beginInsertRows(QtCore.QModelIndex(), row, row)
        # # whatever code
        # self.endInsertRows()

        self.layoutAboutToBeChanged.emit()
        self.layoutChanged.emit()

    def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlag:
        """
        Get the display mode
        :param index:
        :return:
        """
        if self.transposed:
            attr_idx = index.row()
        else:
            attr_idx = index.column()

        if self.editable and self.attributes[attr_idx] not in self.non_editable_attributes:
            return QtCore.Qt.ItemFlag.ItemIsEditable | QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled

    def rowCount(self, parent: QtCore.QModelIndex = None) -> int:
        """
        Get number of rows
        :param parent:
        :return:
        """
        if self.transposed:
            return self.c
        else:
            return self.r

    def columnCount(self, parent: QtCore.QModelIndex = None) -> int:
        """
        Get number of columns
        :param parent:
        :return:
        """
        if self.transposed:
            return self.r
        else:
            return self.c

    def data_raw(self, r, c):
        """
        Get the data to display
        :param r: row index
        :param c: column index
        :return:
        """
        if self.transposed:
            obj_idx = c
            attr_idx = r
        else:
            obj_idx = r
            attr_idx = c

        prop = self.property_list[attr_idx]

        if prop.tpe is Bus:
            return self.objects[obj_idx].get_value(prop=prop, t_idx=self.time_index_).name
        else:
            return self.objects[obj_idx].get_value(prop=prop, t_idx=self.time_index_)

    def data_with_type(self, index: QtCore.QModelIndex):
        """
        Get the data to display
        :param index:
        :return:
        """
        if self.transposed:
            obj_idx = index.column()
            attr_idx = index.row()
        else:
            obj_idx = index.row()
            attr_idx = index.column()

        if obj_idx < len(self.objects):
            prop = self.property_list[attr_idx]

            if prop.tpe is Bus:
                return self.objects[obj_idx].get_value(prop=prop, t_idx=self.time_index_).name
            else:
                return self.objects[obj_idx].get_value(prop=prop, t_idx=self.time_index_)
        else:
            # there is a mismatch because the element was deleted without refreshing this table model
            return ""

    def data(self, index: QtCore.QModelIndex, role=None):
        """
        Get the data to display
        :param index:
        :param role:
        :return:
        """
        if len(self.objects) == 0:
            return None

        if index.isValid():
            if self.transposed:
                attr_idx = index.row()
            else:
                attr_idx = index.column()

            if role == QtCore.Qt.ItemDataRole.DisplayRole:

                if self.property_list[attr_idx].is_date:
                    dt = QtCore.QDateTime.fromSecsSinceEpoch(self.data_with_type(index))
                    formatted_date = dt.toString("yyyy/MM/dd")
                    return formatted_date
                else:
                    return str(self.data_with_type(index))

            elif role == QtCore.Qt.ItemDataRole.BackgroundRole:

                if self.property_list[attr_idx].is_color:
                    return QtGui.QColor(str(self.data_with_type(index)))

        return None

    def setData(self, index: QtCore.QModelIndex, value: Union[float, str], role: Union[int, None] = None) -> bool:
        """
        Set data by simple editor (whatever text)
        :param index:
        :param value:
        :param role:
        :return:
        """
        if len(self.objects) == 0:
            return True

        if self.transposed:
            obj_idx = index.column()
            attr_idx = index.row()
        else:
            obj_idx = index.row()
            attr_idx = index.column()

        prop = self.property_list[attr_idx]

        # check taken values
        if self.attributes[attr_idx] in self.check_unique:
            taken = self.attr_taken(self.attributes[attr_idx], value)
        else:
            taken = False

        if not taken:
            if obj_idx < len(self.objects):
                if self.attributes[attr_idx] not in self.non_editable_attributes:

                    if prop.tpe is ContingencyGroup and value != "":
                        value2 = ContingencyGroup(value)
                    else:
                        value2 = value

                    self.objects[obj_idx].set_value(prop=prop, t_idx=self.time_index_, value=value2)
                else:
                    pass  # the column cannot be edited
            else:
                # the object was deleted without refreshing this table
                pass

        return True

    def attr_taken(self, attr, val):
        """
        Checks if the attribute value is taken
        :param attr:
        :param val:
        :return:
        """
        for obj in self.objects:
            if val == getattr(obj, attr):
                return True
        return False

    def headerData(self,
                   section: int,
                   orientation: QtCore.Qt.Orientation,
                   role=QtCore.Qt.ItemDataRole.DisplayRole):
        """
        Get the headers to display
        :param section:
        :param orientation:
        :param role:
        :return:
        """
        if len(self.objects) == 0:
            return None

        if self._hide_headers_mode is True:
            return None

        if role == QtCore.Qt.ItemDataRole.DisplayRole:

            if self.transposed:
                # for the properties in the schematic view
                if orientation == QtCore.Qt.Orientation.Horizontal:
                    return self.objects[0].device_type.value if len(self.objects) else 'Value'
                elif orientation == QtCore.Qt.Orientation.Vertical:
                    if self.units[section] != '':
                        return self.attributes[section] + ' [' + self.units[section] + ']'
                    else:
                        return self.attributes[section]
            else:
                # Normal
                if orientation == QtCore.Qt.Orientation.Horizontal:
                    if section < len(self.attributes):
                        if self.units[section] != '':
                            return self.attributes[section] + ' [' + self.units[section] + ']'
                        else:
                            return self.attributes[section]
                    else:
                        return "Deleted :/"

                elif orientation == QtCore.Qt.Orientation.Vertical:
                    if section < len(self.objects):
                        return str(section) + ':' + str(self.objects[section])
                    else:
                        return "Deleted :/"

        # add a tooltip
        if role == QtCore.Qt.ItemDataRole.ToolTipRole:
            if section < self.c:
                if self.units[section] != "":
                    unit = '\nUnits: ' + self.units[section]
                else:
                    unit = ''
                return self.attributes[section] + unit + ' \n' + self.tips[section]
            else:
                # somehow the index is out of range
                return ""

        return None

    def copy_to_column(self, index: QtCore.QModelIndex) -> None:
        """
        Copy the value pointed by the index to all the other cells in the column
        :param index: QModelIndex instance
        """
        value = self.data_with_type(index=index)
        col = index.column()

        for row in range(self.rowCount()):

            if self.transposed:
                obj_idx = col
                attr_idx = row
            else:
                obj_idx = row
                attr_idx = col

            if self.attributes[attr_idx] not in self.non_editable_attributes:
                setattr(self.objects[obj_idx], self.attributes[attr_idx], value)
            else:
                pass  # the column cannot be edited

    def get_data(self):
        """

        :return:
        """
        nrows = self.rowCount()
        ncols = self.columnCount()
        data = np.empty((nrows, ncols), dtype=object)

        for j in range(ncols):
            for i in range(nrows):
                data[i, j] = self.data_raw(r=i, c=j)

        columns = [self.headerData(i, orientation=QtCore.Qt.Orientation.Horizontal,
                                   role=QtCore.Qt.ItemDataRole.DisplayRole) for i in range(ncols)]

        index = [self.headerData(i, orientation=QtCore.Qt.Orientation.Vertical,
                                 role=QtCore.Qt.ItemDataRole.DisplayRole) for i in range(nrows)]

        return index, columns, data

    def copy_to_clipboard(self):
        """

        :return:
        """
        if self.columnCount() > 0:

            index, columns, data = self.get_data()

            data = data.astype(str)

            # header first
            txt = '\t' + '\t'.join(columns) + '\n'

            # data
            for t, index_value in enumerate(index):
                txt += str(index_value) + '\t' + '\t'.join(data[t, :]) + '\n'

            # copy to clipboard
            cb = QtWidgets.QApplication.clipboard()
            cb.clear()
            cb.setText(txt)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import List
from PySide6 import QtWidgets
from VeraGrid.Gui.Analysis.analysis_gui import Ui_MainWindow
from VeraGrid.Gui.Analysis.object_plot_analysis import grid_analysis, GridErrorLog, FixableErrorOutOfRange
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGrid.Gui.general_dialogues import LogsDialogue, Logger


class GridAnalysisGUI(QtWidgets.QMainWindow):
    """
    GridAnalysisGUI
    """
    def __init__(self, circuit: MultiCircuit):
        """

        :param circuit: MultiCircuit
        """
        QtWidgets.QMainWindow.__init__(self)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.setWindowTitle('Grid analysis')

        # set the circuit
        self.circuit = circuit

        self.object_types = [dev.device_type.value for dev in circuit.get_template_objects_list()]

        # declare logs
        self.log = GridErrorLog()

        self.fixable_errors: List[FixableErrorOutOfRange] = []

        self.ui.actionSave_diagnostic.triggered.connect(self.save_diagnostic)
        self.ui.actionAnalyze.triggered.connect(self.analyze_all)
        self.ui.actionFix_issues.triggered.connect(self.fix_all)

        self.analyze_all()

    def analyze_all(self):
        """
        Analyze the model data
        :return:
        """
        self.log = GridErrorLog()

        # declare logs
        self.fixable_errors = grid_analysis(
            circuit=self.circuit,
            analyze_ts=self.ui.fixTimeSeriesCheckBox.isChecked(),
            imbalance_threshold=self.ui.activePowerImbalanceSpinBox.value() / 100.0,
            v_low=self.ui.genVsetMinSpinBox.value(),
            v_high=self.ui.genVsetMaxSpinBox.value(),
            tap_min=self.ui.transformerTapModuleMinSpinBox.value(),
            tap_max=self.ui.transformerTapModuleMaxSpinBox.value(),
            transformer_virtual_tap_tolerance=self.ui.virtualTapToleranceSpinBox.value() / 100.0,
            branch_connection_voltage_tolerance=self.ui.lineNominalVoltageToleranceSpinBox.value() / 100.0,
            min_vcc=self.ui.transformerVccMinSpinBox.value(),
            max_vcc=self.ui.transformerVccMaxSpinBox.value(),
            branch_x_threshold=1e-4,
            condition_number_threshold=1e-4,
            logger=self.log
        )

        # set logs
        self.ui.logsTreeView.setModel(self.log.get_model())

    def fix_all(self):
        """
        Fix all detected fixable errors
        :return:
        """
        logger = Logger()
        for fixable_err in self.fixable_errors:
            fixable_err.fix(logger=logger,
                            fix_ts=self.ui.fixTimeSeriesCheckBox.isChecked())

        if logger.has_logs():
            dlg = LogsDialogue("Fixed issues", logger)
            dlg.setModal(True)
            dlg.exec()

        # re-analyze
        self.analyze_all()

    def save_diagnostic(self):
        """
        save_diagnostic
        :return:
        """
        files_types = "Excel (*.xlsx)"

        fname = 'Grid error analysis.xlsx'

        filename, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'Save file', fname, files_types)

        if filename != '':
            self.log.save(filename)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'psse_import_gui.ui'
##
## Created by: Qt User Interface Compiler version 6.10.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QDialog, QFormLayout,
    QFrame, QPushButton, QSizePolicy, QSpacerItem,
    QVBoxLayout, QWidget)
from VeraGrid.Gui.Icons.icons_rc import *
from VeraGrid.Gui.Icons.icons_rc import *
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_PsseImportDialog(object):
    def setupUi(self, PsseImportDialog):
        if not PsseImportDialog.objectName():
            PsseImportDialog.setObjectName(u"PsseImportDialog")
        PsseImportDialog.resize(318, 202)
        self.verticalLayout_2 = QVBoxLayout(PsseImportDialog)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.frame_77 = QFrame(PsseImportDialog)
        self.frame_77.setObjectName(u"frame_77")
        self.frame_77.setMinimumSize(QSize(300, 0))
        self.frame_77.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_77.setFrameShadow(QFrame.Shadow.Raised)
        self.formLayout = QFormLayout(self.frame_77)
        self.formLayout.setObjectName(u"formLayout")
        self.use_short_names_checkBox = QCheckBox(self.frame_77)
        self.use_short_names_checkBox.setObjectName(u"use_short_names_checkBox")

        self.formLayout.setWidget(1, QFormLayout.ItemRole.SpanningRole, self.use_short_names_checkBox)

        self.adjust_taps_to_discrete_positions_checkBox = QCheckBox(self.frame_77)
        self.adjust_taps_to_discrete_positions_checkBox.setObjectName(u"adjust_taps_to_discrete_positions_checkBox")

        self.formLayout.setWidget(0, QFormLayout.ItemRole.SpanningRole, self.adjust_taps_to_discrete_positions_checkBox)

        self.importButton = QPushButton(self.frame_77)
        self.importButton.setObjectName(u"importButton")

        self.formLayout.setWidget(4, QFormLayout.ItemRole.SpanningRole, self.importButton)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.formLayout.setItem(3, QFormLayout.ItemRole.LabelRole, self.verticalSpacer)

        self.flatten_virtual_taps_checkBox = QCheckBox(self.frame_77)
        self.flatten_virtual_taps_checkBox.setObjectName(u"flatten_virtual_taps_checkBox")

        self.formLayout.setWidget(2, QFormLayout.ItemRole.SpanningRole, self.flatten_virtual_taps_checkBox)


        self.verticalLayout_2.addWidget(self.frame_77)


        self.retranslateUi(PsseImportDialog)

        QMetaObject.connectSlotsByName(PsseImportDialog)
    # setupUi

    def retranslateUi(self, PsseImportDialog):
        PsseImportDialog.setWindowTitle(QCoreApplication.translate("PsseImportDialog", u"PSS/e Import", None))
#if QT_CONFIG(tooltip)
        self.use_short_names_checkBox.setToolTip(QCoreApplication.translate("PsseImportDialog", u"<html><head/><body><p>Use &quot;I_J_ckt&quot; insetad of </p><p>&quot;I_Iname_VI_J_Jname_VJ_ckt&quot;</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.use_short_names_checkBox.setText(QCoreApplication.translate("PsseImportDialog", u"Use short branch names", None))
#if QT_CONFIG(tooltip)
        self.adjust_taps_to_discrete_positions_checkBox.setToolTip(QCoreApplication.translate("PsseImportDialog", u"<html><head/><body><p>PSS/e taps might not come adjusted </p><p>to the specified tap positions, </p><p>Do you want to adjust them?</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.adjust_taps_to_discrete_positions_checkBox.setText(QCoreApplication.translate("PsseImportDialog", u"Adjust taps to integer positios", None))
        self.importButton.setText(QCoreApplication.translate("PsseImportDialog", u"Import", None))
#if QT_CONFIG(tooltip)
        self.flatten_virtual_taps_checkBox.setToolTip(QCoreApplication.translate("PsseImportDialog", u"<html><head/><body><p>PSS/e does not handle the difference of nominal </p><p>voltage between the bus and the transformer (the vitual tap) </p><p>If checked, this will make the transformer nominal voltages </p><p>equal to the buses nominal voltage to behave like PSS/e</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.flatten_virtual_taps_checkBox.setText(QCoreApplication.translate("PsseImportDialog", u"Flatten viatual taps", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import numpy as np
import pandas as pd
from typing import List, Union
from PySide6 import QtCore, QtWidgets
from warnings import warn

from VeraGridEngine.Devices.Parents.editable_device import EditableDevice
from VeraGridEngine.enumerations import DeviceType
from VeraGrid.Gui.gui_functions import (ComboDelegate, TextDelegate, FloatDelegate, ComplexDelegate)
from VeraGrid.Gui.wrappable_table_model import WrappableTableModel


class ObjectHistory:
    """
    ObjectHistory
    """

    def __init__(self, max_undo_states: int = 100) -> None:
        """
        Constructor
        :param max_undo_states: maximum number of undo states
        """
        self.max_undo_states = max_undo_states
        self.position = 0
        self.undo_stack = list()
        self.redo_stack = list()

    def add_state(self, action_name, data: dict):
        """
        Add an undo state
        :param action_name: name of the action that was performed
        :param data: dictionary {column index -> profile array}
        """

        # if the stack is too long delete_with_dialogue the oldest entry
        if len(self.undo_stack) > (self.max_undo_states + 1):
            self.undo_stack.pop(0)

        # stack the newest entry
        self.undo_stack.append((action_name, data))

        self.position = len(self.undo_stack) - 1

        # print('Stored', action_name)

    def redo(self):
        """
        Re-do table
        :return: table instance
        """
        val = self.redo_stack.pop()
        self.undo_stack.append(val)
        return val

    def undo(self):
        """
        Un-do table
        :return: table instance
        """
        val = self.undo_stack.pop()
        self.redo_stack.append(val)
        return val

    def can_redo(self):
        """
        is it possible to redo?
        :return: True / False
        """
        return len(self.redo_stack) > 0

    def can_undo(self):
        """
        Is it possible to undo?
        :return: True / False
        """
        return len(self.undo_stack) > 0


class ProfilesModel(WrappableTableModel):
    """
    Class to populate a Qt table view with profiles from objects
    """

    def __init__(self,
                 time_array: pd.DatetimeIndex,
                 elements: List[EditableDevice],
                 device_type: DeviceType,
                 magnitude: str,
                 data_format,
                 parent,
                 max_undo_states=100):
        """

        :param time_array: array of time
        :param device_type: string with Load, StaticGenerator, etc...
        :param magnitude: magnitude to display 'S', 'P', etc...
        :param data_format:
        :param parent: Parent object: the QTableView object
        :param max_undo_states:
        """
        WrappableTableModel.__init__(self, parent)

        self.parent = parent

        self.data_format = data_format

        self.time_array = time_array

        self.device_type = device_type

        self.magnitude = magnitude

        self.non_editable_indices = list()

        self.editable = True

        self.elements = elements

        self.formatter = lambda x: "%.2f" % x

        # contains copies of the table
        self.history = ObjectHistory(max_undo_states)

        # add the initial state
        # self.add_state(columns=range(self.columnCount()), action_name='initial')

        self.set_delegates()

    def set_delegates(self) -> None:
        """
        Set the cell editor types depending on the attribute_types array
        :return:
        """

        if self.data_format is bool:
            delegate = ComboDelegate(self.parent, [True, False], ['True', 'False'])
            self.parent.setItemDelegate(delegate)

        elif self.data_format is float:
            delegate = FloatDelegate(self.parent)
            self.parent.setItemDelegate(delegate)

        elif self.data_format is str:
            delegate = TextDelegate(self.parent)
            self.parent.setItemDelegate(delegate)

        elif self.data_format is complex:
            delegate = ComplexDelegate(self.parent)
            self.parent.setItemDelegate(delegate)

    def update(self):
        """
        update
        """
        # row = self.rowCount()
        # self.beginInsertRows(QtCore.QModelIndex(), row, row)
        # # whatever code
        # self.endInsertRows()

        self.layoutAboutToBeChanged.emit()

        self.layoutChanged.emit()

    def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlag:
        """
        Get the display mode
        :param index:
        :return:
        """

        if self.editable and index.column() not in self.non_editable_indices:
            return QtCore.Qt.ItemFlag.ItemIsEditable | QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled

    def rowCount(self, parent: QtCore.QModelIndex = None) -> int:
        """
        Get number of rows
        :param parent:
        :return:
        """
        return len(self.time_array)

    def columnCount(self, parent: Union[None, QtCore.QModelIndex] = None) -> int:
        """
        Get number of columns
        :param parent:
        :return:
        """
        return len(self.elements)

    def data(self, index: QtCore.QModelIndex, role: int = QtCore.Qt.ItemDataRole.DisplayRole) -> Union[str, None]:
        """
        Get the data to display
        :param index:
        :param role:
        :return:
        """
        if index.isValid():
            if role == QtCore.Qt.ItemDataRole.DisplayRole:
                c = index.column()
                r = index.row()
                profile_attr_name = self.elements[c].properties_with_profile[self.magnitude]
                profile = getattr(self.elements[c], profile_attr_name)
                return str(profile[r])

        return None

    def setData(self,
                index: QtCore.QModelIndex,
                value: float,
                role: QtCore.Qt.ItemDataRole = QtCore.Qt.ItemDataRole.DisplayRole) -> bool:
        """
        Set data by simple editor (whatever text)
        :param index:
        :param value:
        :param role:
        :return:
        """
        c = index.column()
        if c not in self.non_editable_indices:
            r = index.row()
            profile_attr_name = self.elements[index.column()].properties_with_profile[self.magnitude]
            profile = getattr(self.elements[index.column()], profile_attr_name)
            profile[r] = value

            # self.add_state(columns=[c], action_name='')
        else:
            pass  # the column cannot be edited

        return True

    def headerData(self,
                   section: int,
                   orientation: QtCore.Qt.Orientation,
                   role=QtCore.Qt.ItemDataRole.DisplayRole):
        """
        Get the headers to display
        :param section:
        :param orientation:
        :param role:
        :return:
        """

        if self._hide_headers_mode is True:
            return None

        if role == QtCore.Qt.ItemDataRole.DisplayRole:

            if orientation == QtCore.Qt.Orientation.Horizontal:
                if section < len(self.elements):
                    return str(self.elements[section].name)
            elif orientation == QtCore.Qt.Orientation.Vertical:
                if self.time_array is None:
                    return str(section)
                else:
                    return pd.to_datetime(self.time_array[section]).strftime('%d-%m-%Y %H:%M:%S')

        return None

    def paste_from_clipboard(self, row_idx=0, col_idx=0):
        """

        Args:
            row_idx:
            col_idx:
        """
        n = len(self.elements)
        nt = len(self.time_array)

        if n > 0:
            formatter = self.elements[0].registered_properties[self.magnitude].tpe

            # copy to clipboard
            cb = QtWidgets.QApplication.clipboard()
            text = cb.text()

            rows = text.split('\n')

            mod_cols = list()

            # gather values
            for r, row in enumerate(rows):

                values = row.split('\t')
                r2 = r + row_idx
                for c, val in enumerate(values):

                    c2 = c + col_idx

                    try:
                        val2 = formatter(val)
                        parsed = True
                    except:
                        warn("could not parse '" + str(val) + "'")
                        parsed = False
                        val2 = ''

                    if parsed:
                        if c2 < n and r2 < nt:
                            mod_cols.append(c2)
                            prof = self.elements[c2].get_profile(magnitude=self.magnitude)
                            arr = prof.toarray()
                            arr[r2] = val2
                            prof.set(arr)
                        else:
                            print('Out of profile bounds')

        else:
            # there are no elements
            pass

    def copy_to_clipboard(self, cols: Union[None, List[int]] = None) -> bool:
        """
        Copy profiles to clipboard
        :param cols:
        :return:
        """

        if cols is None:
            elements = self.elements
        else:
            if len(cols) > 0:
                elements = [self.elements[i] for i in cols]
            else:
                elements = self.elements

        n = len(elements)

        if n > 0:

            nt = len(self.time_array)

            # gather values
            names = np.empty(n, dtype=object)
            values = np.empty((nt, n), dtype=object)

            for c in range(n):
                names[c] = elements[c].name
                prof = elements[c].get_profile(self.magnitude)
                values[:, c] = prof.toarray().astype(str)

            # header first
            data = '\t' + '\t'.join(names) + '\n'

            # data
            for t, date in enumerate(self.time_array):
                data += str(date) + '\t' + '\t'.join(values[t, :]) + '\n'

            # copy to clipboard
            cb = QtWidgets.QApplication.clipboard()
            cb.clear()
            cb.setText(data)

            return True
        else:
            # there are no elements
            return False

    def add_state(self, columns: List[int], action_name: str = ''):
        """
        Compile data of an action and store the data in the undo history
        :param columns: list of column indices changed
        :param action_name: name of the action
        :return: None
        """
        data = dict()

        for col in columns:
            # profile_property = self.elements[col].properties_with_profile[self.magnitude]
            # data[col] = getattr(self.elements[col], profile_property).copy()
            data[col] = self.elements[col].get_profile(self.magnitude)
            # TODO: check if devices do not have a profile

        self.history.add_state(action_name, data)

    def restore(self, data: dict):
        """
        Set profiles data from undo history
        :param data: dictionary comming from the history
        :return:
        """
        for col, array in data.items():
            profile_property = self.elements[col].properties_with_profile[self.magnitude]
            setattr(self.elements[col], profile_property, array)

    def undo(self):
        """
        Un-do table changes
        """
        if self.history.can_undo():
            action, data = self.history.undo()

            self.restore(data)

            print('Undo ', action)

            self.update()

    def redo(self):
        """
        Re-do table changes
        """
        if self.history.can_redo():
            action, data = self.history.redo()

            self.restore(data)

            print('Redo ', action)

            self.update()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from VeraGridEngine.enumerations import DeviceType

device_type_icons = {
    "Regions": ":/Icons/icons/map.png",
    DeviceType.CountryDevice.value: ":/Icons/icons/country.png",
    DeviceType.CommunityDevice.value: ":/Icons/icons/community.png",
    DeviceType.RegionDevice.value: ":/Icons/icons/region.png",
    DeviceType.MunicipalityDevice.value: ":/Icons/icons/municipality.png",
    DeviceType.AreaDevice.value: ":/Icons/icons/area.png",
    DeviceType.ZoneDevice.value: ":/Icons/icons/zone.png",

    "Substation": ":/Icons/icons/bus_icon.png",
    DeviceType.SubstationDevice.value: ":/Icons/icons/substation.png",
    DeviceType.VoltageLevelDevice.value: ":/Icons/icons/voltage_level.png",
    DeviceType.BusBarDevice.value: ":/Icons/icons/bus_bar_icon.png",
    DeviceType.BusDevice.value: ":/Icons/icons/bus_icon.png",
    DeviceType.SwitchDevice.value: ":/Icons/icons/switch.png",

    "Injections": ":/Icons/icons/add_load.png",
    DeviceType.GeneratorDevice.value: ":/Icons/icons/gen.png",
    DeviceType.BatteryDevice.value: ":/Icons/icons/batt.png",
    DeviceType.LoadDevice.value: ":/Icons/icons/load_dev.png",
    DeviceType.StaticGeneratorDevice.value: ":/Icons/icons/sta_gen.png",
    DeviceType.ExternalGridDevice.value: ":/Icons/icons/external_grid.png",
    DeviceType.ShuntDevice.value: ":/Icons/icons/shunt.png",
    DeviceType.ControllableShuntDevice.value: ":/Icons/icons/controllable_shunt.png",
    DeviceType.CurrentInjectionDevice.value: ":/Icons/icons/load_dev.png",

    "Branches": ":/Icons/icons/reactance.png",
    DeviceType.LineDevice.value: ":/Icons/icons/ac_line.png",
    DeviceType.DCLineDevice.value: ":/Icons/icons/dc.png",
    DeviceType.Transformer2WDevice.value: ":/Icons/icons/to_transformer.png",
    DeviceType.WindingDevice.value: ":/Icons/icons/winding.png",
    DeviceType.Transformer3WDevice.value: ":/Icons/icons/transformer3w.png",
    DeviceType.SeriesReactanceDevice.value: ":/Icons/icons/reactance.png",
    DeviceType.HVDCLineDevice.value: ":/Icons/icons/to_hvdc.png",
    DeviceType.VscDevice.value: ":/Icons/icons/vsc.png",
    DeviceType.UpfcDevice.value: ":/Icons/icons/upfc.png",

    "Fluid": ":/Icons/icons/dam_gray.png",
    DeviceType.FluidNodeDevice.value: ":/Icons/icons/fluid_node.png",
    DeviceType.FluidPathDevice.value: ":/Icons/icons/fluid_path.png",
    DeviceType.FluidTurbineDevice.value: ":/Icons/icons/fluid_turbine.png",
    DeviceType.FluidPumpDevice.value: ":/Icons/icons/fluid_pump.png",
    DeviceType.FluidP2XDevice.value: ":/Icons/icons/fluid_p2x.png",

    "Groups": ":/Icons/icons/groups.png",
    DeviceType.BranchGroupDevice.value: ":/Icons/icons/branch_group.png",
    DeviceType.ModellingAuthority.value: ":/Icons/icons/modelling_authority.png",
    DeviceType.FacilityDevice.value: ":/Icons/icons/powerplant.png",

    "Contingencies": ":/Icons/icons/contingency_group.png",
    DeviceType.ContingencyGroupDevice.value: ":/Icons/icons/contingency_group.png",
    DeviceType.ContingencyDevice.value: ":/Icons/icons/contingency.png",
    DeviceType.RemedialActionGroupDevice.value: ":/Icons/icons/remedial_action_group.png",
    DeviceType.RemedialActionDevice.value: ":/Icons/icons/remedial_action.png",
    DeviceType.ShortCircuitEvent.value: ":/Icons/icons/contingency.png",

    "Investments": ":/Icons/icons/investment_group.png",
    DeviceType.InvestmentsGroupDevice.value: ":/Icons/icons/investment_group.png",
    DeviceType.InvestmentDevice.value: ":/Icons/icons/investment_dev.png",

    "Dynamic": ":/Icons/icons/dyn_gray.png",
    DeviceType.RmsEventsGroupDevice.value: ":/Icons/icons/dyn_gray.png",
    DeviceType.RmsEventDevice.value: ":/Icons/icons/dyn_gray.png",
    DeviceType.EmtEventsGroupDevice.value: ":/Icons/icons/dyn_gray.png",
    DeviceType.EmtEventDevice.value: ":/Icons/icons/dyn_gray.png",

    "Associations": ":/Icons/icons/associations.png",
    DeviceType.Technology.value: ":/Icons/icons/technology.png",
    DeviceType.FuelDevice.value: ":/Icons/icons/fuel.png",
    DeviceType.EmissionGasDevice.value: ":/Icons/icons/emission.png",
    DeviceType.Owner.value: ":/Icons/icons/owner.png",

    "Measurements": ":/Icons/icons/measurement.png",
    DeviceType.PiMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.QiMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.PfMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.QfMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.PtMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.QtMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.PgMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.QgMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.IfMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.ItMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.VmMeasurementDevice.value: ":/Icons/icons/measurement.png",
    DeviceType.VaMeasurementDevice.value: ":/Icons/icons/measurement.png",

    "Catalogue": ":/Icons/icons/Catalogue.png",
    DeviceType.WireDevice.value: ":/Icons/icons/ac_line.png",
    DeviceType.OverheadLineTypeDevice.value: ":/Icons/icons/tower.png",
    DeviceType.UnderGroundLineDevice.value: ":/Icons/icons/ac_line.png",
    DeviceType.SequenceLineDevice.value: ":/Icons/icons/ac_line.png",
    DeviceType.TransformerTypeDevice.value: ":/Icons/icons/to_transformer.png",
    DeviceType.RmsModelTemplateDevice.value: ":/Icons/icons/dyn_gray.png",
    DeviceType.EmtModelTemplateDevice.value: ":/Icons/icons/dyn_gray.png",
    DeviceType.DynamicModelHostDevice.value: ":/Icons/icons/dyn_gray.png",
}



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

from PySide6 import QtWidgets
from PySide6.QtCore import Qt
from PySide6.QtGui import QColor
from typing import List, Dict, Any

from VeraGridEngine.Devices.Parents.editable_device import GCProp, GCPROP_TYPES
from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.enumerations import DeviceType, VoltageLevelTypes, ActionType
from VeraGridEngine.Devices.types import ALL_DEV_TYPES


def populate_tree(tree_widget, base: MultiCircuit, diff: MultiCircuit,
                  all_elms_base_dict: Dict[str, ALL_DEV_TYPES], diff_creation: bool = False):
    """
    Build the display tree
    :param tree_widget:
    :param base:
    :param diff:
    :param all_elms_base_dict:
    :param diff_creation:
    :return:
    """
    tree_widget.clear()
    logger = Logger()

    diff_name = diff.name

    nt = base.get_time_number()
    detailed_profile_comparison = False

    custom_red = QColor(173, 0, 43)
    custom_green = QColor(0, 173, 61)
    custom_blue = QColor(0, 93, 173)

    # all no_action
    all_no_action = True
    # ["Grid", "Object type", "action", "idtag", "name", "property", "value", "new value"]

    for elm in diff.items():

        if elm.action == ActionType.Add:
            parent_item = QtWidgets.QTreeWidgetItem(
                tree_widget,
                [f"{diff_name}", f"{elm.device_type.value}", "Add", f"{elm.idtag}", f"{elm.name}", "", "", ""]
            )
            parent_item.setCheckState(0, Qt.CheckState.Checked)
            parent_item.setBackground(2, custom_green)
            all_no_action = False

        elif elm.action == ActionType.Delete:
            parent_item = QtWidgets.QTreeWidgetItem(
                tree_widget,
                [f"{diff_name}", f"{elm.device_type.value}", "Del", f"{elm.idtag}", f"{elm.name}", "", "", ""]
            )
            parent_item.setCheckState(0, Qt.CheckState.Checked)
            parent_item.setBackground(2, custom_red)
            all_no_action = False

        elif elm.action == ActionType.NoAction:
            parent_item = QtWidgets.QTreeWidgetItem(
                tree_widget,
                [f"{diff_name}", f"{elm.device_type.value}", "-", f"{elm.idtag}", f"{elm.name}", "", "", ""]
            )
            # parent_item.setCheckState(0, Qt.CheckState.Checked)
            # parent_item.setBackground(2, custom_green)

        elif elm.action == ActionType.Modify:
            parent_item = QtWidgets.QTreeWidgetItem(
                tree_widget,
                [f"{diff_name}", f"{elm.device_type.value}", "Mod", f"{elm.idtag}", f"{elm.name}", "", "", ""]
            )
            parent_item.setCheckState(0, Qt.CheckState.Checked)
            parent_item.setBackground(2, custom_blue)
            all_no_action = False

            elm_from_base = all_elms_base_dict.get(elm.idtag, None)

            if elm_from_base is not None:

                if diff_creation:
                    action, changed_props = elm.compare(
                        other=elm_from_base,
                        logger=logger,
                        detailed_profile_comparison=detailed_profile_comparison,
                        nt=nt
                    )

                else:
                    action = ActionType.Modify
                    changed_props_keys = elm.get_all_diff_changes_dict()
                    changed_props = [elm.registered_properties[prop] for prop in changed_props_keys]

                for prop in changed_props:
                    base_val = elm_from_base.get_property_value(prop=prop, t_idx=None)
                    new_val = elm.get_property_value(prop=prop, t_idx=None)

                    elm.set_diff_change(property_name=prop.name, selected=True)

                    child_item = QtWidgets.QTreeWidgetItem(
                        parent_item,
                        [f"{diff_name}", f"{elm.device_type.value}", "Mod", f"{elm.idtag}", 
                         f"{elm.name}", prop.name, str(base_val), str(new_val)]
                    )
                    child_item.setCheckState(0, Qt.CheckState.Checked)

    return all_no_action



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0
import sys
from typing import Union
import numpy as np
import pandas as pd
from PySide6.QtWidgets import QApplication
from PySide6 import QtCore, QtWidgets

from VeraGrid.Gui.SystemScaler.system_scaler_ui import Ui_Dialog
from VeraGrid.Gui.gui_functions import get_list_model, FloatDelegate
from VeraGrid.Gui.messages import yes_no_question
from VeraGridEngine.enumerations import DeviceType
from VeraGridEngine.Devices.multi_circuit import MultiCircuit


class SystemScalingModel(QtCore.QAbstractTableModel):
    """
    Class to populate a Qt table view with a pandas data frame
    """

    def __init__(self, device_tpe: DeviceType, grid: MultiCircuit, parent: QtWidgets.QTableView):
        """

        :param device_tpe:
        :param grid:
        :param parent:
        """
        QtCore.QAbstractTableModel.__init__(self, parent)
        self.parent = parent
        self.device_tpe = device_tpe
        self.grid = grid
        self.objects = self.grid.get_elements_by_type(device_type=device_tpe)
        self.injections_per_type = self.grid.get_injection_devices_grouped_by_group_type(group_type=device_tpe)

        self._cols = ['Load factor', 'Generation factor', 'Total load (MW)', 'Total generation (MW)']
        self._editable = [True, True, False, False]
        self._index = [elm.name for elm in self.objects]
        self.r = len(self._index)
        self.c = len(self._cols)
        self._data = np.ones((self.r, self.c + 2))
        self.isDate = False

        if len(self._index) > 0:
            if isinstance(self._index[0], np.datetime64):
                self._index = pd.to_datetime(self._index)
                self.isDate = True

        self.formatter = lambda x: "%.2f" % x
        self.set_delegates()

        self.original_powers = np.zeros((self.r, 2))

        # compute totals per type
        for i in range(self.r):
            gens = self.injections_per_type[i].get(DeviceType.GeneratorDevice, list())
            loads = self.injections_per_type[i].get(DeviceType.LoadDevice, list())

            # get the original area, zone, etc. power
            self.original_powers[i, 0] = sum([elm.P + elm.G + elm.Ii for elm in loads])
            self.original_powers[i, 1] = sum([elm.P for elm in gens])

            # compute the total scaling power of the area, zone, etc...
            self._data[i, 2] = self.original_powers[i, 0] * self._data[i, 0]
            self._data[i, 3] = self.original_powers[i, 1] * self._data[i, 1]

    def flags(self, index: QtCore.QModelIndex):
        """
        Get the display mode
        :param index:
        :return:
        """

        if self._editable[index.column()]:
            return (QtCore.Qt.ItemFlag.ItemIsEditable |
                    QtCore.Qt.ItemFlag.ItemIsEnabled |
                    QtCore.Qt.ItemFlag.ItemIsSelectable)
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled

    def update(self):
        """
        update table
        """
        # row = self.rowCount()
        # self.beginInsertRows(QtCore.QModelIndex(), row, row)
        # # whatever code
        # self.endInsertRows()
        self.layoutAboutToBeChanged.emit()
        self.layoutChanged.emit()

    def set_delegates(self) -> None:
        """
        Set the cell editor types depending on the attribute_types array
        """

        for i in range(self.c):
            delegate = FloatDelegate(self.parent)
            self.parent.setItemDelegateForColumn(i, delegate)

    def rowCount(self, parent: Union[QtCore.QModelIndex, None] = None):
        """

        :param parent:
        :return:
        """
        return self.r

    def columnCount(self, parent: Union[QtCore.QModelIndex, None] = None):
        """

        :param parent:
        :return:
        """
        return self.c

    def data(self, index, role=QtCore.Qt.ItemDataRole.DisplayRole):
        """

        :param index:
        :param role:
        :return:
        """
        if index.isValid():
            if role == QtCore.Qt.ItemDataRole.DisplayRole:
                return self.formatter(self._data[index.row(), index.column()])
                # return str(self._data[index.row(), index.column()])
        return None

    def headerData(self,
                   section: int,
                   orientation: QtCore.Qt.Orientation,
                   role=QtCore.Qt.ItemDataRole.DisplayRole):
        """

        :param section:
        :param orientation:
        :param role:
        :return:
        """
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if orientation == QtCore.Qt.Orientation.Horizontal:
                return self._cols[section]
            elif orientation == QtCore.Qt.Orientation.Vertical:
                if self._index is None:
                    return section
                else:
                    if self.isDate:
                        return self._index[section].strftime('%Y/%m/%d  %H:%M.%S')
                    else:
                        return str(self._index[section])
        return None

    def setData(self, index, value, role=QtCore.Qt.ItemDataRole.DisplayRole):
        """

        :param index:
        :param value:
        :param role:
        :return:
        """
        if self._editable[index.column()]:
            if value != "":
                i = index.row()
                self._data[i, index.column()] = value

                # update the total scaling power of the area, zone, etc...
                self._data[i, 2] = self.original_powers[i, 0] * self._data[i, 0]
                self._data[i, 3] = self.original_powers[i, 1] * self._data[i, 1]

        return True

    def apply_scaling(self, with_time_series: bool = False):
        """
        Aply the scaling to the objects
        :param with_time_series: scale time profiles too?
        """
        for i in range(self.r):
            gens = self.injections_per_type[i].get(DeviceType.GeneratorDevice, list())
            loads = self.injections_per_type[i].get(DeviceType.LoadDevice, list())
            load_scale = self._data[i, 0]
            gen_scale = self._data[i, 1]

            for elm in loads:
                elm.P *= load_scale
                elm.Q *= load_scale
                elm.G *= load_scale
                elm.B *= load_scale
                elm.Ii *= load_scale
                elm.Ir *= load_scale

                if with_time_series:
                    elm.P_prof *= load_scale
                    elm.Q_prof *= load_scale
                    elm.G_prof *= load_scale
                    elm.B_prof *= load_scale
                    elm.Ii_prof *= load_scale
                    elm.Ir_prof *= load_scale

            for elm in gens:
                elm.P *= gen_scale

                if with_time_series:
                    elm.P_prof *= gen_scale


class SystemScaler(QtWidgets.QDialog):
    """
    SystemScaler GUI
    """

    def __init__(self, grid: MultiCircuit, parent=None):
        """

        :param parent:
        """
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.setWindowTitle('System scaling')

        self.grid = grid
        self.data_model: Union[SystemScalingModel, None] = None

        self.groups = [DeviceType.AreaDevice,
                       DeviceType.ZoneDevice,
                       DeviceType.CountryDevice,
                       DeviceType.RegionDevice,
                       DeviceType.CommunityDevice,
                       DeviceType.MunicipalityDevice,
                       DeviceType.SubstationDevice]

        self.ui.group_combo_box.setModel(get_list_model([elm.value for elm in self.groups]))
        self.ui.group_combo_box.setCurrentIndex(0)
        self.group_change()

        self.ui.doit_button.clicked.connect(self.do_it)
        self.ui.group_combo_box.currentIndexChanged.connect(self.group_change)

    def group_change(self) -> None:
        """

        :return:
        """
        if self.ui.group_combo_box.currentIndex() > -1:
            txt = self.ui.group_combo_box.currentText()

            for group in self.groups:
                if txt == group.value:
                    self.data_model = SystemScalingModel(device_tpe=group,
                                                         grid=self.grid,
                                                         parent=self.ui.groups_scaling_table_view)
                    self.ui.groups_scaling_table_view.setModel(self.data_model)

    def do_it(self):
        """

        :return:
        """
        ok = yes_no_question("This operation will alter the generation "
                             "and load composition irreversibly\nAre you sure?",
                             "System scaling")

        if ok:
            if self.data_model:
                self.data_model.apply_scaling(with_time_series=self.ui.aply_to_time_series_check_box.isChecked())

            self.close()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = SystemScaler(grid=MultiCircuit())
    # window.resize(1.61 * 700.0, 600.0)  # golden ratio
    window.show()
    sys.exit(app.exec())



import os
import sys
import subprocess
import pathlib
import shutil


def convert_svgs(folder: str, backup: bool = True):
    """
    Convert all SVG files in a folder to 'plain SVG' using Inkscape CLI.

    Args:
        folder: Path to the folder containing SVG files.
        backup: Whether to keep a backup of the original files (.bak).
    """
    folder_path = pathlib.Path(folder)

    if not folder_path.exists():
        print(f"❌ Folder not found: {folder}")
        return

    for svg_file in folder_path.rglob("*.svg"):
        out_file = svg_file.with_suffix(".tmp.svg")

        print(f"Converting {svg_file} -> {out_file}")
        try:
            subprocess.run(
                [
                    "inkscape",
                    str(svg_file),
                    "--export-type=svg",
                    f"--export-filename={out_file}"
                ],
                check=True
            )

            if backup:
                backup_file = svg_file.with_suffix(svg_file.suffix + ".bak")
                shutil.move(svg_file, backup_file)
                print(f"  Backup saved: {backup_file}")

            shutil.move(out_file, svg_file)
            print(f"  ✅ Replaced {svg_file} with plain SVG")

        except subprocess.CalledProcessError as e:
            print(f"  ❌ Failed to convert {svg_file}: {e}")
        except Exception as e:
            print(f"  ⚠️ Error with {svg_file}: {e}")


def convert_to_png(root_folder: str, size: int = 64, overwrite: bool = False):
    """
    Recursively converts all SVG files in a folder (and subfolders) to PNGs using Inkscape.
    - root_folder: starting directory
    - size: output width and height in pixels
    - overwrite: overwrite existing PNGs if True
    """
    # Locate inkscape
    inkscape_exe = os.environ.get("INKSCAPE", "inkscape")
    try:
        subprocess.run([inkscape_exe, "--version"], check=True, capture_output=True)
    except Exception:
        if sys.platform == "darwin":
            print("❌ Inkscape not found. On macOS try:")
            print("   export INKSCAPE='/Applications/Inkscape.app/Contents/MacOS/inkscape'")
        else:
            print("❌ Inkscape not found. Add it to PATH or set $INKSCAPE.")
        return

    total, converted, failed = 0, 0, 0

    for dirpath, _, filenames in os.walk(root_folder):
        for fname in filenames:
            if not fname.lower().endswith(".svg"):
                continue
            total += 1
            svg_path = os.path.join(dirpath, fname)
            png_path = os.path.splitext(svg_path)[0] + ".png"

            if os.path.exists(png_path) and not overwrite:
                print(f"⏭️  Skipping existing {png_path}")
                continue

            print(f"🖼️  {svg_path} -> {png_path}")
            try:
                subprocess.run(
                    [
                        inkscape_exe,
                        svg_path,
                        "--export-type=png",
                        f"--export-filename={png_path}",
                        f"--export-width={size}",
                        f"--export-height={size}",
                    ],
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
                print(f"  ✅ Converted")
                converted += 1
            except subprocess.CalledProcessError:
                print(f"  ❌ Failed")
                failed += 1

    print(f"\n--- Summary ---")
    print(f"Total SVGs   : {total}")
    print(f"Converted PNG: {converted}")
    print(f"Failed       : {failed}")


if __name__ == "__main__":
    dir_path = os.path.dirname(os.path.realpath(__file__))
    # convert_svgs("icons")
    convert_to_png(dir_path)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import sys
from typing import List, TYPE_CHECKING
import numpy as np
from PySide6 import QtWidgets

from VeraGrid.Gui.Diagrams.MapWidget.Substation.substation_graphic_item import SubstationGraphicItem
from VeraGrid.Gui.Diagrams.MapWidget.grid_map_widget import haversine_distance
from VeraGrid.Gui.ProceduralGrid.procedural_grid_ui import Ui_Dialog
from VeraGrid.Gui.gui_functions import get_list_model, enums_to_model
from VeraGridEngine.enumerations import ProceduralGridMethods

if TYPE_CHECKING:
    from VeraGrid.Gui.Main.SubClasses.simulations import SimulationsMain


class ProceduralGridWindow(QtWidgets.QDialog):
    """
    SystemScaler GUI
    """

    def __init__(self, app: SimulationsMain, parent=None):
        """

        :param parent:
        """
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Procedural grid expansion')

        # Setup combobox
        self.methods_dict, methods_mdl = enums_to_model(
            [ProceduralGridMethods.SteinerAlone,
             ProceduralGridMethods.SteinerAndOptimization]
        )
        self.ui.methodComboBox.setModel(methods_mdl)
        self.ui.methodComboBox.setCurrentIndex(0)

        # setup candidates

        # setup connection substations
        self.connection_substations_graphics_list: List[SubstationGraphicItem] = list()
        self.candidate_list: List[SubstationGraphicItem] = list()
        self.all_substations_graphics_in_diagram_list: List[SubstationGraphicItem] = list()
        for se_idx, se_api_object, se_graphic in app.get_current_diagram_substations():

            self.all_substations_graphics_in_diagram_list.append(se_graphic)

            if se_graphic.isSelected():
                self.connection_substations_graphics_list.append(se_graphic)

        self.ui.targetSubstationListView.setModel(get_list_model(
            [se_graphic.api_object for se_graphic in self.connection_substations_graphics_list]
        ))

        # Button click
        self.ui.computeButton.clicked.connect(self.compute)
        self.ui.computeCandidatesButton.clicked.connect(self.compute_candidates)

    def compute_candidates(self):
        """

        :return:
        """
        n_number = len(self.connection_substations_graphics_list)
        mc_number = len(self.all_substations_graphics_in_diagram_list)
        top_n = self.ui.topClosestSpinBox.value()

        distances = np.zeros((mc_number, n_number))

        for j, se_graphic_target in enumerate(self.connection_substations_graphics_list):
            for i, se_graphic_candidate in enumerate(self.all_substations_graphics_in_diagram_list):

                distances[i, j] = haversine_distance(
                    lat1=se_graphic_candidate.lat,
                    lon1=se_graphic_candidate.lon,
                    lat2=se_graphic_target.lat,
                    lon2=se_graphic_target.lon
                )

        if distances.shape[0] > 0:

            distances2 = np.min(distances, axis=1)
            sorted_indices = np.argsort(distances2, axis=0)

            self.candidate_list.clear()
            for i in sorted_indices[:top_n]:
                self.candidate_list.append(self.all_substations_graphics_in_diagram_list[i])

            # display the candidates list
            self.ui.candidateSubstationListView.setModel(get_list_model(
                [se_graphic.api_object for se_graphic in self.candidate_list]
            ))

    def compute(self):
        """
        Computation
        :return:
        """
        method: ProceduralGridMethods = self.methods_dict[self.ui.methodComboBox.currentText()]


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from PySide6.QtWidgets import (QDialog, QVBoxLayout, QHBoxLayout, QLabel, QComboBox, QDialogButtonBox)
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Devices.Substation.substation import Substation


class BusSelectorDialogue(QDialog):
    """
    NewMapLineDialogue
    """
    def __init__(self, grid: MultiCircuit, se: Substation, parent=None):
        """
        Constructor
        :param grid: MultiCircuit
        :param se: Substation
        :param parent: ?
        """
        QDialog.__init__(self, parent)
        self.setWindowTitle("Bus selection")

        # Create layout
        main_layout = QVBoxLayout()

        # Create horizontal layout for labels and combo-boxes
        combo_layout = QHBoxLayout()

        self.buses = grid.get_substation_buses(substation=se)
        self._is_valid = len(self.buses) > 0

        # Create labels and combo-boxes
        label1 = QLabel(se.name)
        self.combo_box = QComboBox()
        self.combo_box.addItems([bus.name for bus in self.buses])
        if len(self.buses) > 0:
            self.combo_box.setCurrentIndex(0)

        # Create vertical layout for first label and combobox
        vbox1 = QVBoxLayout()
        vbox1.addWidget(label1)
        vbox1.addWidget(self.combo_box)

        # Add vertical layouts to horizontal layout
        combo_layout.addLayout(vbox1)

        # Add horizontal layout to main layout
        main_layout.addLayout(combo_layout)

        # Create and add buttons
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        # Align the button box to the bottom-right
        button_layout = QHBoxLayout()
        button_layout.addStretch(1)
        button_layout.addWidget(button_box)

        # Add button layout to main layout
        main_layout.addLayout(button_layout)

        # Set main layout
        self.setLayout(main_layout)

    def is_valid(self):
        """

        :return:
        """
        return self._is_valid

    def bus(self):
        """

        :return:
        """
        idx = self.combo_box.currentIndex()
        if idx > -1:
            return self.buses[idx]
        else:
            return None

    def accept(self):
        """
        accept
        """
        super().accept()

    def reject(self):
        """
        reject
        """
        self._is_valid = False
        super().reject()


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

import numpy as np

from typing import Union, List
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QPushButton, QVBoxLayout, QDialog, QLabel, QDoubleSpinBox, QComboBox, QCheckBox
from VeraGrid.Gui.gui_functions import get_list_model
from VeraGrid.Gui.messages import error_msg, warning_msg, yes_no_question
from VeraGridEngine.Devices.Branches.line import Line, SequenceLineType, OverheadLineType, UndergroundLineType
from VeraGridEngine.Devices.multi_circuit import MultiCircuit


class LineEditor(QDialog):
    """
    LineEditor
    """

    def __init__(self,
                 line: Line,
                 grid: MultiCircuit):
        """
        Line Editor constructor
        :param line: Branch object to update
        :param grid: MultiCircuit
        """
        super(LineEditor, self).__init__()

        # keep pointer to the line object
        self.line = line

        self.Sbase = grid.Sbase

        self.frequency = grid.fBase

        Vnom = line.get_max_bus_nominal_voltage()
        self.templates = list()

        for lst in [grid.sequence_line_types,
                    grid.underground_cable_types,
                    grid.overhead_line_types]:
            for temp in lst:
                if Vnom == temp.Vnom:
                    self.templates.append(temp)

        self.current_template = line.template

        self.selected_template = None

        self.setObjectName("self")

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.NoContextMenu)

        self.layout = QVBoxLayout(self)

        # ------------------------------------------------------------------------------------------
        # Set the object values
        # ------------------------------------------------------------------------------------------

        Vf = self.line.bus_from.Vnom
        # Vt = self.line.bus_to.Vnom

        if Vf <= 0.0:
            error_msg(f"Vnom in bus {self.line.bus_from} is {Vf}\n"
                      f"That causes an infinite base admittance.\n"
                      f"The process has been aborted.\n"
                      f"Please correct the data and try again.",
                      title="Line editor initialization")
            return

        Zbase = (Vf * Vf) / self.Sbase
        Ybase = 1 / Zbase
        length = self.line.length

        if length == 0:
            length = 1.0

        r_ohm = self.line.R * Zbase / length
        x_ohm = self.line.X * Zbase / length
        b_us = self.line.B * Ybase / length * 1e6
        I_KA = np.round(self.line.rate / (Vf * 1.73205080757), 6)  # current in kA

        # ------------------------------------------------------------------------------------------

        # catalogue
        self.catalogue_combo = QComboBox()
        if len(self.templates) > 0:
            self.catalogue_combo.setModel(get_list_model(self.templates))

            if self.current_template is not None:
                try:
                    idx = self.templates.index(self.current_template)
                    self.catalogue_combo.setCurrentIndex(idx)

                    if isinstance(self.current_template, SequenceLineType):
                        I_KA = self.current_template.Imax
                        r_ohm = self.current_template.R
                        x_ohm = self.current_template.X
                        b_us = self.current_template.B

                    if isinstance(self.current_template, UndergroundLineType):
                        I_KA = self.current_template.Imax
                        r_ohm = self.current_template.R
                        x_ohm = self.current_template.X
                        b_us = self.current_template.B

                    elif isinstance(self.current_template, OverheadLineType):

                        if self.current_template.check():
                            R1, X1, Bsh1, I_kA = self.current_template.get_sequence_values(
                                circuit_idx=self.line.circuit_idx,
                                seq=1
                            )
                            r_ohm = R1
                            x_ohm = X1
                            b_us = Bsh1
                        else:
                            warning_msg(text=f"The template {self.current_template.name} contains errors",
                                        title="Load template")

                except ValueError:
                    pass

        # load template
        self.load_template_btn = QPushButton()
        self.load_template_btn.setText('Load template values')
        self.load_template_btn.clicked.connect(self.load_template_btn_click)

        # line length
        self.l_spinner = QDoubleSpinBox()
        self.l_spinner.setMinimum(0)
        self.l_spinner.setMaximum(9999999)
        self.l_spinner.setDecimals(6)
        self.l_spinner.setValue(length)
        self.l_spinner.setSuffix(' km')

        # Max current
        self.i_spinner = QDoubleSpinBox()
        self.i_spinner.setMinimum(0)
        self.i_spinner.setMaximum(9999999)
        self.i_spinner.setDecimals(6)
        self.i_spinner.setValue(I_KA)
        self.i_spinner.setSuffix(' kA')

        # R
        self.r_spinner = QDoubleSpinBox()
        self.r_spinner.setMinimum(0)
        self.r_spinner.setMaximum(9999999)
        self.r_spinner.setDecimals(6)
        self.r_spinner.setValue(r_ohm)
        self.r_spinner.setSuffix(' Ω/km')

        # X
        self.x_spinner = QDoubleSpinBox()
        self.x_spinner.setMinimum(0)
        self.x_spinner.setMaximum(9999999)
        self.x_spinner.setDecimals(6)
        self.x_spinner.setValue(x_ohm)
        self.x_spinner.setSuffix(' Ω/km')

        # B
        self.b_spinner = QDoubleSpinBox()
        self.b_spinner.setMinimum(0)
        self.b_spinner.setMaximum(9999999)
        self.b_spinner.setDecimals(6)
        self.b_spinner.setValue(b_us)
        self.b_spinner.setSuffix(" uS/Km")

        # Circuit Index
        max_circuits = 0
        if isinstance(self.current_template, OverheadLineType):
            max_circuits = self.current_template.n_circuits
        self.circuit_idx = QDoubleSpinBox()
        self.circuit_idx.setMinimum(1)
        self.circuit_idx.setMaximum(max_circuits)
        self.circuit_idx.setDecimals(0)
        self.circuit_idx.setValue(float(self.line.circuit_idx))
        self.circuit_idx.setSuffix("")

        # apply to profile
        self.apply_to_profile = QCheckBox()
        self.apply_to_profile.setToolTip("Apply the newly computed values like the rating to the profile")
        self.apply_to_profile.setChecked(True)
        self.apply_to_profile.setText("Apply to profiles")

        # accept button
        self.accept_btn = QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        # add all to the GUI
        self.layout.addWidget(QLabel("Available templates"))
        self.layout.addWidget(self.catalogue_combo)
        self.catalogue_combo.currentIndexChanged.connect(self.update_max_circuits)
        self.layout.addWidget((QLabel("Circuit index:")))
        self.layout.addWidget(self.circuit_idx)
        self.layout.addWidget(self.load_template_btn)
        self.layout.addWidget(QLabel(""))

        self.layout.addWidget(QLabel("L: Line length"))
        self.layout.addWidget(self.l_spinner)

        self.layout.addWidget(QLabel("Imax: Max. current @" + str(int(Vf)) + " [kV]"))
        self.layout.addWidget(self.i_spinner)

        self.layout.addWidget(QLabel("R: Resistance"))
        self.layout.addWidget(self.r_spinner)

        self.layout.addWidget(QLabel("X: Inductance"))
        self.layout.addWidget(self.x_spinner)

        self.layout.addWidget(QLabel("S: Susceptance"))
        self.layout.addWidget(self.b_spinner)

        self.layout.addWidget(self.apply_to_profile)

        self.layout.addWidget(self.accept_btn)

        self.setLayout(self.layout)

        self.setWindowTitle('Line editor')

    def accept_click(self) -> None:
        """
        Set the values
        :return:
        """

        length = self.l_spinner.value()

        if length != 0.0:

            if self.selected_template is not None:
                self.line.disable_auto_updates()
                self.line.set_length(val=length)
                self.line.set_circuit_idx(val=int(self.circuit_idx.value()), obj=self.selected_template)
                template = self.selected_template
                self.line.apply_template(obj=template, Sbase=self.Sbase, freq=self.frequency)
                self.line.enable_auto_updates()
                self.accept()
            else:
                response = yes_no_question(
                    text="Warning: You did not load template values. The circuit index will not be updated. "
                         "Line parameters will be based on the provided values for Length, Max Current, Resistance, "
                         "Reactance, and Susceptance.\n\n"
                         "Do you want to continue without a template?",
                    title="No Template Selected"
                )

                if response:  # User selected Yes
                    wf = 2 * np.pi * self.frequency
                    self.line.fill_design_properties(
                        r_ohm=self.r_spinner.value(),  # ohm / km
                        x_ohm=self.x_spinner.value(),  # ohm / km
                        c_nf=self.b_spinner.value() * 1e3 / wf,  # nF / km
                        length=length,  # km
                        Imax=self.i_spinner.value(),  # kA
                        freq=self.frequency,  # Hz
                        Sbase=self.Sbase,  # MVA
                        apply_to_profile=self.apply_to_profile.isChecked()
                    )
                    self.accept()
                else:
                    # User canceled, do nothing
                    return
        else:
            error_msg(text="The length cannot be 0!", title="Accept line design values")

    def load_template(self, template: Union[SequenceLineType, OverheadLineType, UndergroundLineType]):
        """
        Load a template in the editor
        :param template: line compatible template
        :return:
        """

        if isinstance(template, SequenceLineType):
            self.i_spinner.setValue(template.Imax)
            self.r_spinner.setValue(template.R)
            self.x_spinner.setValue(template.X)
            self.b_spinner.setValue(template.B)

            self.selected_template = template

        elif isinstance(template, UndergroundLineType):
            self.i_spinner.setValue(template.Imax)
            self.r_spinner.setValue(template.R)
            self.x_spinner.setValue(template.X)
            self.b_spinner.setValue(template.B)

            self.selected_template = template

        elif isinstance(template, OverheadLineType):
            if template.check():
                R1, X1, Bsh1, I_kA = template.get_sequence_values(
                    circuit_idx=int(self.circuit_idx.value()),
                    seq=1
                )
                self.i_spinner.setValue(I_kA)
                self.r_spinner.setValue(R1)
                self.x_spinner.setValue(X1)
                self.b_spinner.setValue(Bsh1)
                self.circuit_idx.setMaximum(template.n_circuits)
            else:
                warning_msg(text=f"The template {template.name} contains errors",
                            title="Load template")

            self.selected_template = template

    def load_template_btn_click(self):
        """
        Accept template values
        """

        if self.templates is not None:
            idx = self.catalogue_combo.currentIndex()

            if idx > -1:
                self.load_template(template=self.templates[idx])

    def update_max_circuits(self):
        """
        Update the maximum number of circuits
        :return:
        """

        if self.templates is not None:
            idx = self.catalogue_combo.currentIndex()

            if idx > -1:
                template = self.templates[idx]

                if isinstance(template, OverheadLineType):
                    self.circuit_idx.setMaximum(template.n_circuits)
                else:
                    self.circuit_idx.setMaximum(1)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'procedural_grid_ui.ui'
##
## Created by: Qt User Interface Compiler version 6.10.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDialog, QFrame,
    QGridLayout, QHBoxLayout, QLabel, QListView,
    QPushButton, QSizePolicy, QSpacerItem, QSpinBox,
    QWidget)
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(450, 451)
        self.gridLayout = QGridLayout(Dialog)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(1, 1, 1, 1)
        self.label_4 = QLabel(Dialog)
        self.label_4.setObjectName(u"label_4")

        self.gridLayout.addWidget(self.label_4, 4, 0, 1, 1)

        self.targetSubstationListView = QListView(Dialog)
        self.targetSubstationListView.setObjectName(u"targetSubstationListView")

        self.gridLayout.addWidget(self.targetSubstationListView, 6, 0, 1, 3)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 9, 1, 1, 1)

        self.frame = QFrame(Dialog)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.NoFrame)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(1, 1, 1, 1)
        self.topClosestSpinBox = QSpinBox(self.frame)
        self.topClosestSpinBox.setObjectName(u"topClosestSpinBox")
        self.topClosestSpinBox.setMinimum(1)
        self.topClosestSpinBox.setMaximum(9999999)
        self.topClosestSpinBox.setValue(5)

        self.horizontalLayout.addWidget(self.topClosestSpinBox)

        self.computeCandidatesButton = QPushButton(self.frame)
        self.computeCandidatesButton.setObjectName(u"computeCandidatesButton")

        self.horizontalLayout.addWidget(self.computeCandidatesButton)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.computeButton = QPushButton(self.frame)
        self.computeButton.setObjectName(u"computeButton")

        self.horizontalLayout.addWidget(self.computeButton)


        self.gridLayout.addWidget(self.frame, 10, 0, 1, 3)

        self.methodComboBox = QComboBox(Dialog)
        self.methodComboBox.setObjectName(u"methodComboBox")

        self.gridLayout.addWidget(self.methodComboBox, 1, 1, 1, 2)

        self.label_5 = QLabel(Dialog)
        self.label_5.setObjectName(u"label_5")

        self.gridLayout.addWidget(self.label_5, 0, 0, 1, 1)

        self.label = QLabel(Dialog)
        self.label.setObjectName(u"label")

        self.gridLayout.addWidget(self.label, 1, 0, 1, 1)

        self.candidateSubstationListView = QListView(Dialog)
        self.candidateSubstationListView.setObjectName(u"candidateSubstationListView")

        self.gridLayout.addWidget(self.candidateSubstationListView, 8, 0, 1, 3)

        self.label_2 = QLabel(Dialog)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout.addWidget(self.label_2, 7, 0, 1, 3)

        self.label_3 = QLabel(Dialog)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout.addWidget(self.label_3, 5, 0, 1, 3)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Procedural Grid", None))
        self.label_4.setText("")
        self.topClosestSpinBox.setSuffix(QCoreApplication.translate("Dialog", u" Substations", None))
#if QT_CONFIG(tooltip)
        self.computeCandidatesButton.setToolTip(QCoreApplication.translate("Dialog", u"Compute the N closest candiate substations", None))
#endif // QT_CONFIG(tooltip)
        self.computeCandidatesButton.setText(QCoreApplication.translate("Dialog", u"Get candiates", None))
        self.computeButton.setText(QCoreApplication.translate("Dialog", u"Compute", None))
        self.label_5.setText("")
        self.label.setText(QCoreApplication.translate("Dialog", u"Procedural grid method", None))
#if QT_CONFIG(tooltip)
        self.candidateSubstationListView.setToolTip(QCoreApplication.translate("Dialog", u"List of subations that can be connected to the selected substations", None))
#endif // QT_CONFIG(tooltip)
        self.label_2.setText(QCoreApplication.translate("Dialog", u"Connection substation candidates", None))
        self.label_3.setText(QCoreApplication.translate("Dialog", u"Substation to connect", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
import numpy as np
from typing import Dict, List, Union, Any, Tuple, TYPE_CHECKING, Set, Sequence
from PySide6 import QtCore, QtWidgets, QtGui
from collections import defaultdict, OrderedDict

from VeraGridEngine.basic_structures import IntVec
from VeraGridEngine.data_logger import DataLogger
from VeraGridEngine.IO.cim.cgmes.cgmes_circuit import CgmesCircuit, CGMES_ASSETS
from VeraGridEngine.Devices.Branches.line_locations import LineLocations
from VeraGridEngine.Devices.types import ALL_DEV_TYPES
from VeraGrid.Gui.font_config import MENU_FONT_SIZE

if TYPE_CHECKING:
    from VeraGrid.Gui.object_model import ObjectsModel


class TreeDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QComboBox in every
    cell of the column to which it's applied
    """
    commitData = QtCore.Signal(object)
    """
    
    """

    def __init__(self, parent, data=None):
        """
        Constructor
        :param parent: QTableView parent object
        :param data: dictionary of lists
        """
        QtWidgets.QItemDelegate.__init__(self, parent)

        # dictionary of lists
        self.data = data if data is not None else defaultdict()

    @QtCore.Slot()
    def double_click(self):
        """
        double click
        """
        self.commitData.emit(self.sender())

    def createEditor(self, parent, option, index):
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        tree = QtWidgets.QTreeView(parent)

        model = QtGui.QStandardItemModel()
        model.setHorizontalHeaderLabels(['Template'])

        for key in self.data.keys():
            # add parent node
            parent1 = QtGui.QStandardItem(str(key))

            # add children to parent
            for elm in self.data[key]:
                child1 = QtGui.QStandardItem(str(elm))
                parent1.appendRow([child1])

            model.appendRow(parent1)

        tree.setModel(model)
        tree.doubleClicked.connect(self.double_click)
        return tree

    def setEditorData(self, editor, index):
        """

        :param editor:
        :param index:
        """
        print(editor)
        print(index)

    def setModelData(self, editor, model, index):
        """

        :param editor:
        :param model:
        :param index:
        """
        print(editor)
        print(model)
        print(index)

        # model.setData(index, self.object_names[editor.currentIndex()])


class ComboDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QComboBox in every
    cell of the column to which it's applied
    """
    commitData = QtCore.Signal(object)

    def __init__(self, parent: QtWidgets.QTableView, objects: List[Any], object_names: List[str]) -> None:
        """
        Constructor
        :param parent: QTableView parent object
        :param objects: List of objects to set. i.e. [True, False] or [Line1, Line2, ...]
        :param object_names: List of Object names to display. i.e. ['True', 'False']
        """
        QtWidgets.QItemDelegate.__init__(self, parent)

        # objects to sent to the model associated to the combobox. i.e. [True, False]
        self.objects = objects

        # object description to display in the combobox. i.e. ['True', 'False']
        self.object_names = object_names

    @QtCore.Slot()
    def currentIndexChanged(self) -> None:
        """
        currentIndexChanged
        """
        self.commitData.emit(self.sender())

    def createEditor(self, parent, option, index: QtCore.QModelIndex):
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        combo = QtWidgets.QComboBox(parent)
        combo.addItems(self.object_names)
        combo.currentIndexChanged.connect(self.currentIndexChanged)
        return combo

    def setEditorData(self, editor: QtWidgets.QComboBox, index: QtCore.QModelIndex):
        """

        :param editor:
        :param index:
        """
        editor.blockSignals(True)
        val = index.model().data(index, role=QtCore.Qt.ItemDataRole.DisplayRole)
        try:
            idx = self.object_names.index(val)
            editor.setCurrentIndex(idx)
            editor.blockSignals(False)
        except ValueError:
            pass

    def setModelData(self,
                     editor: QtWidgets.QComboBox,
                     model: QtCore.QAbstractItemModel,
                     index: QtCore.QModelIndex):
        """

        :param editor:
        :param model:
        :param index:
        """
        if len(self.objects) > 0:
            if editor.currentIndex() < len(self.objects):
                model.setData(index, self.objects[editor.currentIndex()])


class TextDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QLineEdit in every
    cell of the column to which it's applied
    """

    commitData = QtCore.Signal(object)

    def __init__(self, parent: QtWidgets.QTableView) -> None:
        """
        Constructor
        :param parent: QTableView parent object
        """
        QtWidgets.QItemDelegate.__init__(self, parent)

    @QtCore.Slot()
    def returnPressed(self):
        """
        returnPressed
        """
        self.commitData.emit(self.sender())

    def createEditor(self, parent: QtWidgets.QWidget,
                     option: QtWidgets.QStyleOptionViewItem,
                     index: QtCore.QModelIndex) -> QtWidgets.QLineEdit:
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        editor = QtWidgets.QLineEdit(parent)
        editor.returnPressed.connect(self.returnPressed)
        return editor

    def setEditorData(self, editor: QtWidgets.QLineEdit, index):
        """

        :param editor:
        :param index:
        """
        editor.blockSignals(True)
        val = index.model().data(index, role=QtCore.Qt.ItemDataRole.DisplayRole)
        editor.setText(val)
        editor.blockSignals(False)

    def setModelData(self, editor: QtWidgets.QLineEdit, model, index):
        """

        :param editor:
        :param model:
        :param index:
        """
        model.setData(index, editor.text())


class FloatDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QDoubleSpinBox in every
    cell of the column to which it's applied
    """

    commitData = QtCore.Signal(object)

    def __init__(self,
                 parent: QtWidgets.QTableView,
                 min_: float = -1e200,
                 max_: float = 1e200,
                 decimals: int = 6) -> None:
        """
        Constructor
        :param parent: QTableView parent object
        """
        QtWidgets.QItemDelegate.__init__(self, parent)
        self.min = min_
        self.max = max_
        self.decimals = decimals

    @QtCore.Slot()
    def returnPressed(self) -> None:
        """
        returnPressed
        """
        self.commitData.emit(self.sender())

    def createEditor(self, parent: QtWidgets.QWidget,
                     option: QtWidgets.QStyleOptionViewItem,
                     index: QtCore.QModelIndex) -> QtWidgets.QDoubleSpinBox:
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        editor = QtWidgets.QDoubleSpinBox(parent)
        editor.setMaximum(self.max)
        editor.setMinimum(self.min)
        editor.setDecimals(self.decimals)
        editor.editingFinished.connect(self.returnPressed)
        return editor

    def setEditorData(self, editor: QtWidgets.QDoubleSpinBox, index):
        """

        :param editor:
        :param index:
        """
        editor.blockSignals(True)
        try:
            val = float(index.model().data(index, role=QtCore.Qt.ItemDataRole.DisplayRole))
        except ValueError:
            val = 0.0
        editor.setValue(val)
        editor.blockSignals(False)

    def setModelData(self, editor: QtWidgets.QDoubleSpinBox, model, index):
        """

        :param editor:
        :param model:
        :param index:
        """
        model.setData(index, editor.value())


class IntDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QDoubleSpinBox in every
    cell of the column to which it's applied
    """

    commitData = QtCore.Signal(object)

    def __init__(self, parent: QtWidgets.QTableView, min_: int = -99999, max_: int = 99999) -> None:
        """
        Constructor
        :param parent: QTableView parent object
        """
        QtWidgets.QItemDelegate.__init__(self, parent)
        self.min = min_
        self.max = max_

    @QtCore.Slot()
    def returnPressed(self):
        """
        returnPressed
        """
        self.commitData.emit(self.sender())

    def createEditor(self, parent, option, index):
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        editor = QtWidgets.QSpinBox(parent)
        editor.setMaximum(self.max)
        editor.setMinimum(self.min)
        editor.editingFinished.connect(self.returnPressed)
        return editor

    def setEditorData(self, editor: QtWidgets.QDoubleSpinBox, index):
        """

        :param editor:
        :param index:
        """
        editor.blockSignals(True)
        val = int(index.model().data(index, role=QtCore.Qt.ItemDataRole.DisplayRole))
        editor.setValue(val)
        editor.blockSignals(False)

    def setModelData(self, editor: QtWidgets.QDoubleSpinBox, model, index):
        """

        :param editor:
        :param model:
        :param index:
        """
        model.setData(index, editor.value())


class ComplexDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning Complex Editor in every
    cell of the column to which it's applied
    """

    commitData = QtCore.Signal(object)

    def __init__(self, parent):
        """
        Constructor
        :param parent: QTableView parent object
        """
        QtWidgets.QItemDelegate.__init__(self, parent)

    @QtCore.Slot()
    def returnPressed(self):
        """

        :return:
        """
        self.commitData.emit(self.sender())

    def createEditor(self, parent, option, index):
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        editor = QtWidgets.QFrame(parent)
        main_layout = QtWidgets.QHBoxLayout(editor)
        main_layout.layout().setContentsMargins(0, 0, 0, 0)

        real = QtWidgets.QDoubleSpinBox()
        real.setMaximum(9999)
        real.setMinimum(-9999)
        real.setDecimals(8)

        imag = QtWidgets.QDoubleSpinBox()
        imag.setMaximum(9999)
        imag.setMinimum(-9999)
        imag.setDecimals(8)

        main_layout.addWidget(real)
        main_layout.addWidget(imag)
        # main_layout.addWidget(button)

        # button.clicked.connect(self.returnPressed)

        return editor

    def setEditorData(self, editor: QtWidgets.QFrame, index):
        """

        :param editor:
        :param index:
        :return:
        """
        editor.blockSignals(True)
        val = complex(index.model().data(index, role=QtCore.Qt.ItemDataRole.DisplayRole))
        editor.children()[1].setValue(val.real)
        editor.children()[2].setValue(val.imag)
        editor.blockSignals(False)

    def setModelData(self, editor: QtWidgets.QFrame, model, index):
        """

        :param editor:
        :param model:
        :param index:
        :return:
        """
        val = complex(editor.children()[1].value(), editor.children()[2].value())
        model.setData(index, val)


class LineLocationsDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning LineLocations Editor in every
    cell of the column to which it's applied
    """

    commitData = QtCore.Signal(object)

    def __init__(self, parent):
        """
        Constructor
        :param parent: QTableView parent object
        """
        QtWidgets.QItemDelegate.__init__(self, parent)

        self.line_locations: Union[None, LineLocations] = None

    @QtCore.Slot()
    def returnPressed(self):
        """

        :return:
        """
        self.commitData.emit(self.sender())

    def createEditor(self, parent: QtWidgets.QWidget,
                     option: QtWidgets.QStyleOptionViewItem,
                     index: QtCore.QModelIndex):
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        editor = QtWidgets.QFrame(parent)
        main_layout = QtWidgets.QHBoxLayout(editor)
        main_layout.layout().setContentsMargins(0, 0, 0, 0)

        table = QtWidgets.QTableView()

        main_layout.addWidget(table)
        # main_layout.addWidget(button)

        # button.clicked.connect(self.returnPressed)
        editor.showNormal()

        return editor

    def setEditorData(self, editor: QtWidgets.QFrame, index):
        """

        :param editor:
        :param index:
        :return:
        """
        editor.blockSignals(True)
        self.line_locations: LineLocations = index.model().data(index, role=QtCore.Qt.ItemDataRole.DisplayRole)
        # editor.children()[1].setValue(val.real)
        editor.blockSignals(False)

    def setModelData(self, editor: QtWidgets.QFrame, model: ObjectsModel, index):
        """

        :param editor:
        :param model:
        :param index:
        :return:
        """
        table = editor.children()[1]
        # model.setData(index, val)
        print()


class ColorPickerDelegate(QtWidgets.QItemDelegate):
    """
    Color picker delegate
    """
    commitData = QtCore.Signal(object)

    def __init__(self, parent):
        """

        :param parent:
        """
        super(ColorPickerDelegate, self).__init__(parent)

    @QtCore.Slot()
    def returnPressed(self):
        """

        :return:
        """
        self.commitData.emit(self.sender())

    def createEditor(self, parent, option, index):
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        colorDialog = QtWidgets.QColorDialog(parent)
        return colorDialog

    def setEditorData(self, editor: QtWidgets.QColorDialog, index):
        """

        :param editor:
        :param index:
        """
        editor.blockSignals(True)
        val = index.model().data(index, role=QtCore.Qt.ItemDataRole.DisplayRole)
        color = QtGui.QColor.fromString(val)
        editor.setCurrentColor(color)
        editor.blockSignals(False)

    def setModelData(self, editor: QtWidgets.QColorDialog, model, index):
        """

        :param editor:
        :param model:
        :param index:
        :return:
        """
        model.setData(index, editor.currentColor().name())


class DateTimeDelegate(QtWidgets.QItemDelegate):
    """
    DateTime picker delegate that handles Unix epoch seconds
    and can be initialized with a default epoch date.
    """
    commitData = QtCore.Signal(object)

    def __init__(self, parent, default_epoch: int = 0):
        super().__init__(parent)
        # Store default epoch or None
        self.default_epoch: int = default_epoch

    @QtCore.Slot()
    def returnPressed(self):
        """

        """
        self.commitData.emit(self.sender())

    def createEditor(self, parent, option, index):
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        editor = QtWidgets.QDateTimeEdit(parent)
        editor.setCalendarPopup(True)
        editor.setDisplayFormat("yyyy/MM/dd HH:mm:ss")

        # Set initial datetime from default_epoch if provided
        if self.default_epoch is not None:
            dt = QtCore.QDateTime.fromSecsSinceEpoch(int(self.default_epoch))
            editor.setDateTime(dt)

        editor.dateTimeChanged.connect(lambda: self.commitData.emit(editor))
        return editor

    def setEditorData(self, editor: QtWidgets.QDateTimeEdit, index):
        """

        :param editor:
        :param index:
        """
        editor.blockSignals(True)
        val = index.model().data(index, QtCore.Qt.ItemDataRole.DisplayRole)
        if isinstance(val, (int, float)):
            dt = QtCore.QDateTime.fromSecsSinceEpoch(int(val))
        elif isinstance(val, str):
            dt = QtCore.QDateTime.fromString(val, "yyyy/MM/dd")
        else:
            # fallback to current datetime or default_epoch if provided
            if self.default_epoch is not None:
                dt = QtCore.QDateTime.fromSecsSinceEpoch(int(self.default_epoch))
            else:
                dt = QtCore.QDateTime.currentDateTime()
        editor.setDateTime(dt)
        editor.blockSignals(False)

    def setModelData(self, editor: QtWidgets.QDateTimeEdit, model, index):
        """

        :param editor:
        :param model:
        :param index:
        """
        dt = editor.dateTime()
        epoch_seconds = dt.toSecsSinceEpoch()
        model.setData(index, epoch_seconds)


def get_list_model(lst: Sequence[Union[str, ALL_DEV_TYPES]],
                   checks=False,
                   check_value=False) -> QtGui.QStandardItemModel:
    """
    Pass a list to a list model
    """
    list_model = QtGui.QStandardItemModel()
    if lst is not None:
        if not checks:
            for val in lst:
                # for the list model
                item = QtGui.QStandardItem(str(val))
                item.setEditable(False)
                list_model.appendRow(item)
        else:
            for val in lst:
                # for the list model
                item = QtGui.QStandardItem(str(val))
                item.setEditable(False)
                item.setCheckable(True)
                if check_value:
                    item.setCheckState(QtCore.Qt.CheckState.Checked)
                list_model.appendRow(item)

    return list_model


def get_elm_chck_list_model(lst: List[ALL_DEV_TYPES], check_status: IntVec) -> QtGui.QStandardItemModel:
    """
    Pass a list to a list model
    """
    list_model = QtGui.QStandardItemModel()
    for elm, val in zip(lst, check_status):
        # for the list model
        item = QtGui.QStandardItem(str(elm.name))
        item.setEditable(False)
        item.setCheckable(True)
        if val:
            item.setCheckState(QtCore.Qt.CheckState.Checked)
        list_model.appendRow(item)

    return list_model


def get_chck_list_model(lst: List[str], check_status: List[bool]) -> QtGui.QStandardItemModel:
    """
    Pass a list to a list model
    """
    list_model = QtGui.QStandardItemModel()
    for elm, val in zip(lst, check_status):
        # for the list model
        item = QtGui.QStandardItem(elm)
        item.setEditable(False)
        item.setCheckable(True)
        if val:
            item.setCheckState(QtCore.Qt.CheckState.Checked)
        list_model.appendRow(item)

    return list_model


def enums_to_model(enums_lst: List[Any]) -> Tuple[Dict[str, Any], QtGui.QStandardItemModel]:
    """
    Get the model and dict from a list of Enum value
    :param enums_lst:
    :return: dictionary, model
    """
    d = OrderedDict()
    val_list = list()
    for e in enums_lst:
        d[e.value] = e
        val_list.append(e.value)

    mdl = get_list_model(val_list)

    return d, mdl


class CustomFileSystemModel(QtWidgets.QFileSystemModel):
    """
    CustomFileSystemModel
    """

    def __init__(self, root_path: str, ext_filter: Union[None, List[str]] = None):
        super(CustomFileSystemModel, self).__init__()

        self.ext_filter = ext_filter if ext_filter is not None else ['*.py']

        self.setNameFilters(ext_filter)
        self.setRootPath(root_path)


def get_logger_tree_model(logger: DataLogger) -> QtGui.QStandardItemModel:
    """
    Fill logger tree
    :param logger: Logger instance
    :return: QStandardItemModel instance
    """
    d = logger.to_dict()
    editable = False
    model = QtGui.QStandardItemModel()
    model.setHorizontalHeaderLabels(['Time', 'Element', 'Class', 'Property', 'Value', 'Expected value', 'comment'])
    parent = model.invisibleRootItem()

    for severity, messages_dict in d.items():
        severity_child = QtGui.QStandardItem(severity)

        # print(severity)

        for message, data_list in messages_dict.items():
            message_child = QtGui.QStandardItem(message)

            # print('\t', message)

            for time, elm, elm_class, elm_property, value, expected_value, comment in data_list:
                # print('\t', '\t', time, elm, value, expected_value)

                time_child = QtGui.QStandardItem(time)
                time_child.setEditable(editable)

                elm_child = QtGui.QStandardItem(str(elm))
                elm_child.setEditable(editable)

                elm_class_child = QtGui.QStandardItem(str(elm_class))
                elm_class_child.setEditable(editable)

                elm_property_child = QtGui.QStandardItem(str(elm_property))
                elm_property_child.setEditable(editable)

                value_child = QtGui.QStandardItem(str(value))
                value_child.setEditable(editable)

                expected_val_child = QtGui.QStandardItem(str(expected_value))
                expected_val_child.setEditable(editable)

                comment_val_child = QtGui.QStandardItem(str(comment))
                comment_val_child.setEditable(editable)

                message_child.appendRow([time_child, elm_child, elm_class_child,
                                         elm_property_child, value_child, expected_val_child, comment_val_child])

            message_child.setEditable(editable)

            severity_child.appendRow(message_child)

        severity_child.setEditable(editable)
        parent.appendRow(severity_child)

    return model


def get_icon_list_model(lst: List[Tuple[str, str]], checks=False,
                        check_value=False) -> QtGui.QStandardItemModel:
    """

    :param lst:
    :param checks:
    :param check_value:
    :return:
    """
    list_model = QtGui.QStandardItemModel()
    if lst is not None:
        if not checks:
            for val, icon_path in lst:
                # for the list model
                item = QtGui.QStandardItem(str(val))
                item.setEditable(False)
                icon = QtGui.QIcon()
                icon.addPixmap(QtGui.QPixmap(icon_path))
                item.setIcon(icon)
                list_model.appendRow(item)
        else:
            for val, icon_path in lst:
                # for the list model
                item = QtGui.QStandardItem(str(val))
                icon = QtGui.QIcon()
                icon.addPixmap(QtGui.QPixmap(icon_path))
                item.setIcon(icon)
                item.setEditable(False)
                item.setCheckable(True)
                if check_value:
                    item.setCheckState(QtCore.Qt.CheckState.Checked)
                list_model.appendRow(item)

    return list_model


def get_checked_indices(mdl: QtGui.QStandardItemModel) -> IntVec:
    """
    Get a list of the selected indices in a QStandardItemModel
    :param mdl:
    :return:
    """
    idx = list()
    for row in range(mdl.rowCount()):
        item = mdl.item(row)
        if item.checkState() == QtCore.Qt.CheckState.Checked:
            idx.append(row)

    return np.array(idx)


def get_checked_values(mdl: QtGui.QStandardItemModel) -> List[str]:
    """
    Get a list of the selected values in a QStandardItemModel
    :param mdl:
    :return:
    """
    idx = list()
    for row in range(mdl.rowCount()):
        item = mdl.item(row)
        if item.checkState() == QtCore.Qt.CheckState.Checked:
            idx.append(item.text())

    return idx


def fill_model_from_dict(parent: QtGui.QStandardItem,
                         d: Dict[str, Union[Dict[str, Any], List[str]]],
                         editable=False,
                         icons: Dict[str, str] = None,
                         font_size: int = MENU_FONT_SIZE):
    """
    Fill TreeViewModel from dictionary
    :param parent: Parent QStandardItem
    :param d: item
    :param editable
    :param icons
    :param font_size
    :return: Nothing
    """
    font = QtGui.QFont()
    font.setPointSize(font_size)

    if isinstance(d, dict):
        for k, v in d.items():
            name = str(k)
            child = QtGui.QStandardItem(name)
            child.setEditable(editable)
            child.setFont(font)

            if icons is not None:
                icon_path = icons.get(name, None)
                if icon_path is not None:
                    icon_path = icons[name]
                    _icon = QtGui.QIcon()
                    _icon.addPixmap(QtGui.QPixmap(icon_path))
                    child.setIcon(_icon)
                else:
                    pass
                    # print(f"path {name} has no icon: check 'setup_objects_tree()'")

            parent.appendRow(child)
            fill_model_from_dict(parent=child, d=v, icons=icons)
    elif isinstance(d, list):
        for v in d:
            fill_model_from_dict(parent=parent, d=v, icons=icons)
    else:
        name = str(d)
        item = QtGui.QStandardItem(name)
        item.setFont(font)

        if icons is not None:
            icon_path = icons.get(name, None)
            if icon_path is not None:
                icon_path = icons[name]
                _icon = QtGui.QIcon()
                _icon.addPixmap(QtGui.QPixmap(icon_path))
                item.setIcon(_icon)
            else:
                pass
                # print(f"path {name} has no icon: check 'setup_objects_tree()'")
        item.setEditable(editable)
        parent.appendRow(item)


def get_tree_model(d, top='', icons: Dict[str, str] = None) -> QtGui.QStandardItemModel:
    """
    Build a tree model from a dictionary
    :param d: Any dictionary
    :param top: Table header
    :param icons: Dictionary of icons (name of the node, icon to match)
    :return:
    """
    model = QtGui.QStandardItemModel()
    model.setHorizontalHeaderLabels([top])
    fill_model_from_dict(model.invisibleRootItem(), d=d, editable=False, icons=icons)

    return model


def get_tree_item_path(item: QtGui.QStandardItem) -> List[str]:
    """
    Get the path of an item in a tree
    :param item: QStandardItem
    :return: path in a list
    """
    item_parent = item.parent()
    path = [item.text()]
    while item_parent is not None:
        parent_text = item_parent.text()
        path.append(parent_text)
        item_parent = item_parent.parent()
    path.reverse()
    return path


CIM_OBJECT_ROLE = QtCore.Qt.UserRole + 120
CIM_LOADED_ROLE = QtCore.Qt.UserRole + 121


def _get_cim_object_id(obj: Any):
    if hasattr(obj, 'rdfid') and obj.rdfid is not None:
        return obj.rdfid
    return id(obj)


def _get_cim_object_label(class_tag, device):
    if class_tag is not None:
        return class_tag
    if hasattr(device, 'name'):
        if device.name is not None:
            if device.name != '':
                return device.name
    if hasattr(device, 'rdfid'):
        return device.rdfid
    return str(device)


def _collect_cim_ancestor_ids(item: QtGui.QStandardItem) -> Set[Any]:
    ids: Set[Any] = set()
    current = item
    while current is not None:
        obj = current.data(CIM_OBJECT_ROLE)
        if obj is not None:
            ids.add(_get_cim_object_id(obj))
        current = current.parent()
    return ids


def create_cim_object_item(class_tag,
                           device: CGMES_ASSETS,
                           editable=False,
                           lazy=True) -> QtGui.QStandardItem:
    item = QtGui.QStandardItem(_get_cim_object_label(class_tag, device))
    item.setEditable(editable)
    item.setData(device, CIM_OBJECT_ROLE)
    item.setData(False, CIM_LOADED_ROLE)

    if lazy and getattr(device, 'declared_properties', None):
        # Placeholder child to show expand arrow; removed on load.
        item.appendRow(QtGui.QStandardItem(""))

    if not lazy:
        populate_cim_item_children(item, editable=editable)

    return item


def populate_cim_item_children(item: QtGui.QStandardItem, editable=False):
    """
    Populate immediate children for a CIM object item. Designed for lazy loading.
    """
    device = item.data(CIM_OBJECT_ROLE)
    if device is None:
        return

    if item.data(CIM_LOADED_ROLE):
        return

    ancestor_ids = _collect_cim_ancestor_ids(item)
    item.removeRows(0, item.rowCount())

    for property_name, cim_prop in device.declared_properties.items():

        property_value = getattr(device, property_name)

        if hasattr(property_value, 'rdfid'):

            prop_id = _get_cim_object_id(property_value)
            we_are_in_a_recursive_loop = prop_id in ancestor_ids

            if not we_are_in_a_recursive_loop:
                tpe = str(property_value.tpe)
                class_name_child = create_cim_object_item(class_tag=tpe,
                                                          device=property_value,
                                                          editable=editable,
                                                          lazy=True)
                class_name_child.setEditable(editable)

                property_name_child = QtGui.QStandardItem(tpe)
                property_name_child.setEditable(editable)

                value_child = QtGui.QStandardItem(property_value.rdfid)
                value_child.setEditable(editable)
            else:
                class_name_child = QtGui.QStandardItem("Recursive object (" + str(len(ancestor_ids)) + ")")
                class_name_child.setEditable(editable)

                property_name_child = QtGui.QStandardItem(property_name)
                property_name_child.setEditable(editable)

                value_child = QtGui.QStandardItem(str(property_value))
                value_child.setEditable(editable)
        else:
            tpe = (str(type(property_value)).replace('class', '')
                   .replace("'", "")
                   .replace("<", "")
                   .replace(">", "").strip())

            class_name_child = QtGui.QStandardItem(tpe)
            class_name_child.setEditable(editable)

            property_name_child = QtGui.QStandardItem(property_name)
            property_name_child.setEditable(editable)

            value_child = QtGui.QStandardItem(str(property_value))
            value_child.setEditable(editable)

        item.appendRow([class_name_child, property_name_child, value_child])

    item.setData(True, CIM_LOADED_ROLE)


def add_cim_object_node(class_tag,
                        device: CGMES_ASSETS,
                        editable=False,
                        already_visited: Union[Set, None] = None):
    """

    :param class_tag:
    :param device:
    :param editable:
    :param already_visited:
    :return:
    """
    if already_visited is None:
        already_visited = set()

    if class_tag is None:
        class_tag = _get_cim_object_label(class_tag=None, device=device)

    # create root node
    device_child = QtGui.QStandardItem(class_tag)

    # register visit to avoid cyclic recursion
    device_id = _get_cim_object_id(device)
    already_visited.add(device_id)

    try:
        for property_name, cim_prop in device.declared_properties.items():

            property_value = getattr(device, property_name)

            if hasattr(property_value, 'rdfid'):

                prop_id = _get_cim_object_id(property_value)
                we_are_in_a_recursive_loop = prop_id in already_visited

                if not we_are_in_a_recursive_loop:

                    # if the property is an object, recursively add it
                    tpe = str(property_value.tpe)
                    class_name_child = add_cim_object_node(class_tag=tpe,
                                                           device=property_value,
                                                           editable=editable,
                                                           already_visited=already_visited)
                    class_name_child.setEditable(editable)

                    property_name_child = QtGui.QStandardItem(tpe)
                    property_name_child.setEditable(editable)

                    value_child = QtGui.QStandardItem(property_value.rdfid)
                    value_child.setEditable(editable)
                else:
                    # print('Recursive loop...')
                    # return device_child
                    class_name_child = QtGui.QStandardItem("Recursive object (" + str(len(already_visited)) + ")")
                    class_name_child.setEditable(editable)

                    property_name_child = QtGui.QStandardItem(property_name)
                    property_name_child.setEditable(editable)

                    value_child = QtGui.QStandardItem(str(property_value))
                    value_child.setEditable(editable)
            else:
                # if the property is a value (float, str, bool, etc.) just add it

                tpe = (str(type(property_value)).replace('class', '')
                       .replace("'", "")
                       .replace("<", "")
                       .replace(">", "").strip())

                class_name_child = QtGui.QStandardItem(tpe)
                class_name_child.setEditable(editable)

                property_name_child = QtGui.QStandardItem(property_name)
                property_name_child.setEditable(editable)

                value_child = QtGui.QStandardItem(str(property_value))
                value_child.setEditable(editable)

            device_child.appendRow([class_name_child, property_name_child, value_child])
    finally:
        already_visited.discard(device_id)

    return device_child


def get_cim_tree_model(cim_model: CgmesCircuit):
    """
    Fill logger tree
    :param cim_model: Logger instance
    :return: QStandardItemModel instance
    """

    editable = False
    model = QtGui.QStandardItemModel()
    model.setHorizontalHeaderLabels(['Object class', 'Property', 'Value'])
    root_node = model.invisibleRootItem()

    for class_name, device_list in cim_model.elements_by_type.items():

        class_child = QtGui.QStandardItem(class_name + " (" + str(len(device_list)) + ")")

        for device in device_list:
            # add device with all it's properties
            device_child = create_cim_object_item(class_tag=None,
                                                  device=device,
                                                  editable=editable,
                                                  lazy=True)

            device_child.setEditable(editable)

            class_child.appendRow(device_child)

        class_child.setEditable(editable)
        root_node.appendRow(class_child)

    return model


def add_sub_menu(menu: QtWidgets.QMenu,
                 text: str,
                 icon_path: str = "",
                 icon_pixmap: QtGui.QPixmap = None, ):
    entry = menu.addMenu(text)

    if icon_pixmap is None:
        if len(icon_path) > 0:
            edit_icon = QtGui.QIcon()
            edit_icon.addPixmap(QtGui.QPixmap(icon_path))
            entry.setIcon(edit_icon)
    else:
        # prefer the icon pixmap if provided
        edit_icon = QtGui.QIcon()
        edit_icon.addPixmap(icon_pixmap)
        entry.setIcon(edit_icon)

    return entry


def add_menu_entry(menu: QtWidgets.QMenu,
                   text: str,
                   icon_path: str = "",
                   icon_pixmap: QtGui.QPixmap = None,
                   function_ptr=None,
                   checkeable=False,
                   checked_value=False,
                   font_size: int = MENU_FONT_SIZE) -> QtGui.QAction:
    """
    Add a context menu entry
    :param menu:
    :param text:
    :param icon_path:
    :param icon_pixmap:
    :param function_ptr:
    :param checkeable:
    :param checked_value:
    :param font_size:
    :return:
    """

    entry: QtGui.QAction = menu.addAction(text)

    font = QtGui.QFont()
    font.setPointSize(font_size)
    entry.setFont(font)

    if checkeable:
        entry.setCheckable(checkeable)
        entry.setChecked(bool(checked_value))

    if icon_pixmap is None:
        if len(icon_path) > 0:
            edit_icon = QtGui.QIcon()
            edit_icon.addPixmap(QtGui.QPixmap(icon_path))
            entry.setIcon(edit_icon)
    else:
        # prefer the icon pixmap if provided
        edit_icon = QtGui.QIcon()
        edit_icon.addPixmap(icon_pixmap)
        entry.setIcon(edit_icon)

    if function_ptr is not None:
        entry.triggered.connect(function_ptr)

    return entry


def create_spinbox(value: float, minimum: float, maximum: float, decimals: int = 4) -> QtWidgets.QDoubleSpinBox:
    """

    :param value:
    :param minimum:
    :param maximum:
    :param decimals:
    :return:
    """
    sn_spinner = QtWidgets.QDoubleSpinBox()
    sn_spinner.setMinimum(minimum)
    sn_spinner.setMaximum(maximum)
    sn_spinner.setDecimals(decimals)
    sn_spinner.setValue(value)
    return sn_spinner


def create_int_spinbox(value: int, minimum: int, maximum: int) -> QtWidgets.QSpinBox:
    """

    :param value:
    :param minimum:
    :param maximum:
    :return:
    """
    sn_spinner = QtWidgets.QSpinBox()
    sn_spinner.setMinimum(minimum)
    sn_spinner.setMaximum(maximum)
    sn_spinner.setValue(value)
    return sn_spinner


class AsyncTask(QtCore.QRunnable):
    def __init__(self, task):
        super().__init__()
        self.task = task

    @QtCore.Slot()
    def run(self):
        self.task()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import List, TYPE_CHECKING
from PySide6 import QtWidgets
from PySide6.QtWidgets import (QVBoxLayout, QPushButton)

from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import (SchematicWidget, make_diagram_from_buses,
                                                                    SchematicDiagram)
from VeraGridEngine.Devices.Substation.substation import Substation, DeviceType
from VeraGridEngine.Devices.Substation.bus import Bus
from VeraGridEngine.Devices.multi_circuit import MultiCircuit

if TYPE_CHECKING:
    from VeraGrid.Gui.Main.VeraGridMain import VeraGridMainGUI


class DiagramBusSelectorDialogue(QtWidgets.QDialog):
    """
    ShortCircuitSelector
    """

    def __init__(self,
                 gui: VeraGridMainGUI,
                 grid: MultiCircuit,
                 substation: Substation):
        """

        :param gui:
        :param grid:
        :param substation:
        """
        super().__init__()
        self.setWindowTitle("Bus selection by diagram")

        self.layout = QVBoxLayout(self)

        selected_buses = grid.get_buses_from_objects(elements=[substation], dtype=DeviceType.SubstationDevice)

        if len(selected_buses):
            diagram = make_diagram_from_buses(circuit=grid,
                                              buses=selected_buses,
                                              name=substation.name + " diagram")
        else:
            diagram = SchematicDiagram()

        self.diagram_widget = SchematicWidget(
            gui=gui,
            circuit=grid,
            diagram=diagram,
            default_bus_voltage=gui.ui.defaultBusVoltageSpinBox.value(),
            time_index=gui.get_diagram_slider_index()
        )

        self.layout.addWidget(self.diagram_widget)

        self.select_button = QPushButton("Select")
        self.diagram_widget.frame1_layout.addWidget(self.select_button)
        self.diagram_widget.library_view.setVisible(False)

        self._selected_buses: List[Bus] = list()

        self.select_button.clicked.connect(self.select)

    def select(self):
        """

        :return:
        """
        for i, bus, _ in self.diagram_widget.get_selected_buses():
            self._selected_buses.append(bus)

        self.close()

    def get_selected_buses(self) -> List[Bus]:
        """
        Get the selected buses
        :return:
        """
        return self._selected_buses



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'cgmes_export_gui.ui'
##
## Created by: Qt User Interface Compiler version 6.10.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QDialog,
    QFrame, QGridLayout, QLabel, QListView,
    QPushButton, QSizePolicy, QVBoxLayout, QWidget)
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_CgmesExportDialog(object):
    def setupUi(self, CgmesExportDialog):
        if not CgmesExportDialog.objectName():
            CgmesExportDialog.setObjectName(u"CgmesExportDialog")
        CgmesExportDialog.resize(425, 569)
        self.verticalLayout_2 = QVBoxLayout(CgmesExportDialog)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.frame_79 = QFrame(CgmesExportDialog)
        self.frame_79.setObjectName(u"frame_79")
        self.frame_79.setMinimumSize(QSize(300, 0))
        self.frame_79.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_79.setFrameShadow(QFrame.Shadow.Raised)
        self.gridLayout_5 = QGridLayout(self.frame_79)
        self.gridLayout_5.setObjectName(u"gridLayout_5")
        self.label_139 = QLabel(self.frame_79)
        self.label_139.setObjectName(u"label_139")

        self.gridLayout_5.addWidget(self.label_139, 11, 0, 1, 3)

        self.cgmes_version_comboBox = QComboBox(self.frame_79)
        self.cgmes_version_comboBox.setObjectName(u"cgmes_version_comboBox")

        self.gridLayout_5.addWidget(self.cgmes_version_comboBox, 3, 0, 1, 3)

        self.line_31 = QFrame(self.frame_79)
        self.line_31.setObjectName(u"line_31")
        palette = QPalette()
        brush = QBrush(QColor(186, 189, 182, 255))
        brush.setStyle(Qt.BrushStyle.SolidPattern)
        palette.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.WindowText, brush)
        palette.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.WindowText, brush)
        brush1 = QBrush(QColor(190, 190, 190, 255))
        brush1.setStyle(Qt.BrushStyle.SolidPattern)
        palette.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.WindowText, brush1)
        self.line_31.setPalette(palette)
        self.line_31.setFrameShadow(QFrame.Shadow.Plain)
        self.line_31.setLineWidth(4)
        self.line_31.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_5.addWidget(self.line_31, 1, 0, 1, 3)

        self.selectCGMESBoundarySetButton = QPushButton(self.frame_79)
        self.selectCGMESBoundarySetButton.setObjectName(u"selectCGMESBoundarySetButton")
        self.selectCGMESBoundarySetButton.setMaximumSize(QSize(80, 16777215))

        self.gridLayout_5.addWidget(self.selectCGMESBoundarySetButton, 9, 2, 1, 1)

        self.cgmes_map_regions_like_raw_checkBox = QCheckBox(self.frame_79)
        self.cgmes_map_regions_like_raw_checkBox.setObjectName(u"cgmes_map_regions_like_raw_checkBox")

        self.gridLayout_5.addWidget(self.cgmes_map_regions_like_raw_checkBox, 14, 0, 1, 3)

        self.cgmes_single_profile_per_file_checkBox = QCheckBox(self.frame_79)
        self.cgmes_single_profile_per_file_checkBox.setObjectName(u"cgmes_single_profile_per_file_checkBox")

        self.gridLayout_5.addWidget(self.cgmes_single_profile_per_file_checkBox, 13, 0, 1, 3)

        self.label_135 = QLabel(self.frame_79)
        self.label_135.setObjectName(u"label_135")

        self.gridLayout_5.addWidget(self.label_135, 2, 0, 1, 3)

        self.label_102 = QLabel(self.frame_79)
        self.label_102.setObjectName(u"label_102")

        self.gridLayout_5.addWidget(self.label_102, 10, 0, 1, 1)

        self.label_134 = QLabel(self.frame_79)
        self.label_134.setObjectName(u"label_134")
        palette1 = QPalette()
        brush2 = QBrush(QColor(85, 87, 83, 255))
        brush2.setStyle(Qt.BrushStyle.SolidPattern)
        palette1.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.WindowText, brush2)
        brush3 = QBrush(QColor(0, 0, 0, 255))
        brush3.setStyle(Qt.BrushStyle.SolidPattern)
        palette1.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.WindowText, brush3)
        palette1.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.WindowText, brush1)
        self.label_134.setPalette(palette1)
        font = QFont()
        font.setPointSize(16)
        self.label_134.setFont(font)

        self.gridLayout_5.addWidget(self.label_134, 0, 1, 1, 2)

        self.label_90 = QLabel(self.frame_79)
        self.label_90.setObjectName(u"label_90")

        self.gridLayout_5.addWidget(self.label_90, 8, 0, 1, 3)

        self.cgmes_dc_as_hvdclines_checkBox = QCheckBox(self.frame_79)
        self.cgmes_dc_as_hvdclines_checkBox.setObjectName(u"cgmes_dc_as_hvdclines_checkBox")

        self.gridLayout_5.addWidget(self.cgmes_dc_as_hvdclines_checkBox, 15, 0, 1, 3)

        self.label_137 = QLabel(self.frame_79)
        self.label_137.setObjectName(u"label_137")
        self.label_137.setMinimumSize(QSize(24, 24))
        self.label_137.setMaximumSize(QSize(24, 24))
        self.label_137.setPixmap(QPixmap(u":/Icons/icons/new2.png"))
        self.label_137.setScaledContents(True)

        self.gridLayout_5.addWidget(self.label_137, 0, 0, 1, 1)

        self.cgmes_profiles_listView = QListView(self.frame_79)
        self.cgmes_profiles_listView.setObjectName(u"cgmes_profiles_listView")

        self.gridLayout_5.addWidget(self.cgmes_profiles_listView, 12, 0, 1, 3)

        self.cgmes_boundary_set_label = QLabel(self.frame_79)
        self.cgmes_boundary_set_label.setObjectName(u"cgmes_boundary_set_label")

        self.gridLayout_5.addWidget(self.cgmes_boundary_set_label, 9, 0, 1, 2)

        self.exportButton = QPushButton(self.frame_79)
        self.exportButton.setObjectName(u"exportButton")

        self.gridLayout_5.addWidget(self.exportButton, 16, 2, 1, 1)


        self.verticalLayout_2.addWidget(self.frame_79)


        self.retranslateUi(CgmesExportDialog)

        QMetaObject.connectSlotsByName(CgmesExportDialog)
    # setupUi

    def retranslateUi(self, CgmesExportDialog):
        CgmesExportDialog.setWindowTitle(QCoreApplication.translate("CgmesExportDialog", u"CGMES Export", None))
        self.label_139.setText(QCoreApplication.translate("CgmesExportDialog", u"Profiles to export", None))
#if QT_CONFIG(tooltip)
        self.selectCGMESBoundarySetButton.setToolTip(QCoreApplication.translate("CgmesExportDialog", u"Select the CGMES boundary set (single zip file)", None))
#endif // QT_CONFIG(tooltip)
        self.selectCGMESBoundarySetButton.setText(QCoreApplication.translate("CgmesExportDialog", u"Select", None))
#if QT_CONFIG(tooltip)
        self.cgmes_map_regions_like_raw_checkBox.setToolTip(QCoreApplication.translate("CgmesExportDialog", u"<html><head/><body><p>If active the CGMEs mapping will be:</p><p>GeographicalRegion &lt;-&gt; Area</p><p>SubGeographicalRegion &lt;-&gt; Zone</p><p>Otherwise:</p><p>GeographicalRegion &lt;-&gt; Country</p><p>SubGeographicalRegion &lt;-&gt; Community</p><p><br/></p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.cgmes_map_regions_like_raw_checkBox.setText(QCoreApplication.translate("CgmesExportDialog", u"Map regions like raw files", None))
        self.cgmes_single_profile_per_file_checkBox.setText(QCoreApplication.translate("CgmesExportDialog", u"One file per profile", None))
        self.label_135.setText(QCoreApplication.translate("CgmesExportDialog", u"Export version", None))
        self.label_102.setText("")
        self.label_134.setText(QCoreApplication.translate("CgmesExportDialog", u"CGMES", None))
        self.label_90.setText(QCoreApplication.translate("CgmesExportDialog", u"Boundary set", None))
#if QT_CONFIG(tooltip)
        self.cgmes_dc_as_hvdclines_checkBox.setToolTip(QCoreApplication.translate("CgmesExportDialog", u"<html><head/><body><p>Converters and DC lines in CGMES are attempted to be converted to the simplified HvdcLine objects in VeraGrid</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.cgmes_dc_as_hvdclines_checkBox.setText(QCoreApplication.translate("CgmesExportDialog", u"Treat DC equipement as HvdcLines", None))
        self.label_137.setText("")
#if QT_CONFIG(tooltip)
        self.cgmes_boundary_set_label.setToolTip(QCoreApplication.translate("CgmesExportDialog", u"Path of the CGMES default boundary set (single zip file)", None))
#endif // QT_CONFIG(tooltip)
        self.cgmes_boundary_set_label.setText(QCoreApplication.translate("CgmesExportDialog", u"...", None))
        self.exportButton.setText(QCoreApplication.translate("CgmesExportDialog", u"Export", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import TYPE_CHECKING
from PySide6 import QtWidgets

from VeraGrid.Gui.Diagrams.generic_graphics import Square
from VeraGridEngine.Devices.Injections.controllable_shunt import ControllableShunt
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.injections_template_graphics import InjectionTemplateGraphicItem
from VeraGrid.Gui.Diagrams.Editors.controllable_shunt_editor import ControllableShuntEditor
from VeraGrid.Gui.gui_functions import add_menu_entry

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class ControllableShuntGraphicItem(InjectionTemplateGraphicItem):
    """
    ExternalGrid graphic item
    """

    def __init__(self, parent, api_obj: ControllableShunt, editor: SchematicWidget):
        """

        :param parent:
        :param api_obj:
        :param editor:
        """
        InjectionTemplateGraphicItem.__init__(self,
                                              parent=parent,
                                              api_obj=api_obj,
                                              editor=editor,
                                              device_type_name='external grid',
                                              w=40,
                                              h=40)
        self.set_glyph(glyph=Square(self, 40, 40, "C", self.update_nexus))

    @property
    def api_object(self) -> ControllableShunt:
        return self._api_object

    def contextMenuEvent(self, event):
        """
        Display context menu
        @param event:
        @return:
        """
        menu = self.get_base_context_menu()
        menu.addSection("Controllable shunt")

        add_menu_entry(menu=menu,
                       text="Editor",
                       function_ptr=self.edit,
                       icon_path=":/Icons/icons/edit.png")

        menu.exec_(event.screenPos())

    def edit(self):
        """
        Call the edit dialogue
        :return:
        """
        dlg = ControllableShuntEditor(api_object=self.api_object)
        if dlg.exec():
            pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import re

from typing import List, Dict, Union

from PySide6 import QtWidgets, QtCore, QtGui

from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Utils.Symbolic.symbolic import Var
from VeraGridEngine.Devices.Aggregation.rms_events_group import RmsEventsGroup



# def parse_float(s: str) -> float:
#     """
#     Parse a float from a string accepting both comma and dot as decimal separators.
#     Handles cases like:
#       - "1.5" -> 1.5
#       - "1,5" -> 1.5
#       - "1.234,56" -> 1234.56  (common European)
#       - "1,234.56" -> 1234.56  (common US)
#     Raises ValueError if it cannot be parsed.
#     """
#     if s is None:
#         raise ValueError("Empty string")
#
#     s = s.strip()
#     if s == "":
#         raise ValueError("Empty string")
#
#     # Remove spaces
#     s = s.replace(" ", "")
#
#     # If there are both '.' and ',' assume the last one is the decimal separator:
#     # common convention: "1.234,56" (dot thousands, comma decimal) -> remove dots, comma->dot
#     # or "1,234.56" (comma thousands, dot decimal) -> remove commas, keep dot.
#     if '.' in s and ',' in s:
#         # decide which occurs last
#         if s.rfind(',') > s.rfind('.'):
#             # comma is decimal separator
#             s = s.replace('.', '')
#             s = s.replace(',', '.')
#         else:
#             # dot is decimal separator
#             s = s.replace(',', '')
#     else:
#         # only comma or only dot or none: normalize comma -> dot
#         s = s.replace(',', '.')
#
#     # Now s should look like a normal Python float literal, maybe with +/-
#     if not re.fullmatch(r'[+-]?\d+(\.\d+)?', s):
#         raise ValueError(f"Not a valid float: {s}")
#
#     return float(s)
#
#

class EventRow:
    """Encapsulates a row in the RMS event table."""

    def __init__(self, table, row, parameters_list, events_groups_list):
        self.table = table
        self.row = row

        # --- Checkbox ---
        checkbox = QtWidgets.QTableWidgetItem()
        checkbox.setFlags(
            QtCore.Qt.ItemFlag.ItemIsUserCheckable |
            QtCore.Qt.ItemFlag.ItemIsEnabled
        )
        checkbox.setCheckState(QtCore.Qt.CheckState.Unchecked)
        table.setItem(row, 0, checkbox)

        # --- Parameter selector ---
        self.param_combo = QtWidgets.QComboBox()
        for var in parameters_list:
            self.param_combo.addItem(var.name, var)
        table.setCellWidget(row, 1, self.param_combo)

        # --- Time spinbox ---
        self.time_spin = QtWidgets.QDoubleSpinBox()
        self.time_spin.setDecimals(4)
        self.time_spin.setRange(0.0, 1e9)
        self.time_spin.setSingleStep(0.1)
        self.time_spin.setSuffix(" s")
        table.setCellWidget(row, 2, self.time_spin)

        # --- Value spinbox ---
        self.value_spin = QtWidgets.QDoubleSpinBox()
        self.value_spin.setDecimals(6)
        self.value_spin.setRange(-1e9, 1e9)
        self.value_spin.setSingleStep(0.01)
        table.setCellWidget(row, 3, self.value_spin)

        # --- Event group selector ---
        self.group_combo = QtWidgets.QComboBox()
        for group in events_groups_list:
            self.group_combo.addItem(group.name, group)
        table.setCellWidget(row, 4, self.group_combo)

    def is_checked(self):
        item = self.table.item(self.row, 0)
        return item and item.checkState() == QtCore.Qt.CheckState.Checked

    def get_data(self):
        """Return validated row data."""
        param = self.param_combo.currentData()
        group = self.group_combo.currentData()

        if param is None or group is None:
            raise ValueError("Missing fields")

        t = self.time_spin.value()
        v = self.value_spin.value()

        return param, t, v, group

    def add_group(self, group):
        """Add a new group option to the combo."""
        self.group_combo.addItem(group.name, group)


class RmsEventDialogue(QtWidgets.QDialog):

    def __init__(
        self,
        circuit: MultiCircuit,
        parameters_list: List[Var],
        target_device_name: str,
        parent=None,
    ):
        super().__init__(parent)

        self.setWindowTitle("RMS Event Editor")
        self.setMinimumWidth(600)

        self.circuit = circuit
        self.parameters_list = parameters_list
        self.target_device_name = target_device_name

        self.rows: List[EventRow] = []

        self.data = {
            "parameters": [],
            "target_times": [],
            "values": [],
            "groups": [],
        }

        # ---- Main Layout ----
        layout = QtWidgets.QVBoxLayout(self)

        # --- Device label ---
        label_device = QtWidgets.QLabel(
            f"<b>Target device:</b> {target_device_name}"
        )
        layout.addWidget(label_device)

        # --- Events Table ---
        self.table = QtWidgets.QTableWidget(0, 5)

        self.table.setHorizontalHeaderLabels(
            ["", "Parameter", "Time", "New Value", "Group"]
        )

        self.table.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.ResizeMode.ResizeToContents
        )

        self.table.horizontalHeader().setStretchLastSection(True)

        self.table.verticalHeader().setVisible(False)

        self.table.setSelectionMode(
            QtWidgets.QAbstractItemView.SelectionMode.NoSelection
        )

        layout.addWidget(self.table)

        # --- Groups controls ---
        groups_layout = QtWidgets.QHBoxLayout()

        self.new_group_btn = QtWidgets.QPushButton("➕ New Event Group")

        groups_layout.addStretch()
        groups_layout.addWidget(self.new_group_btn)

        layout.addLayout(groups_layout)


        # --- Table control buttons ---
        table_button_layout = QtWidgets.QHBoxLayout()

        self.add_row_btn = QtWidgets.QPushButton("➕ Add New Event")
        self.remove_row_btn = QtWidgets.QPushButton("❌ Remove Selected Rows")

        table_button_layout.addWidget(self.add_row_btn)
        table_button_layout.addWidget(self.remove_row_btn)

        layout.addLayout(table_button_layout)

        # --- Bottom buttons ---
        button_layout = QtWidgets.QHBoxLayout()

        self.ok_button = QtWidgets.QPushButton("✅ Add Events")
        self.cancel_button = QtWidgets.QPushButton("Cancel")

        button_layout.addStretch()
        button_layout.addWidget(self.ok_button)
        button_layout.addWidget(self.cancel_button)

        layout.addLayout(button_layout)

        # --- Connections ---
        self.add_row_btn.clicked.connect(self.add_row)
        self.remove_row_btn.clicked.connect(self.remove_checked_rows)
        self.ok_button.clicked.connect(self.accept_dialog)
        self.cancel_button.clicked.connect(self.reject)
        self.new_group_btn.clicked.connect(self.create_event_group)

    def add_row(self):
        """Add a new row to the table."""
        row_index = self.table.rowCount()
        self.table.insertRow(row_index)

        row = EventRow(
            self.table,
            row_index,
            self.parameters_list,
            self.circuit.rms_events_groups,
        )

        self.rows.append(row)

    def remove_checked_rows(self):
        """Remove rows where checkbox is checked."""

        rows_to_remove = [r for r in self.rows if r.is_checked()]

        if not rows_to_remove:
            QtWidgets.QMessageBox.information(
                self,
                "No Rows Selected",
                "Please check at least one row to remove.",
            )
            return

        for row in reversed(rows_to_remove):
            self.table.removeRow(row.row)
            self.rows.remove(row)

        # Reindex rows
        for i, row in enumerate(self.rows):
            row.row = i

    def create_event_group(self):
        """Open dialog to create a new RMS events group."""

        dialog = RmsEventsGroupsDialog(self)

        if dialog.exec():
            name = dialog.get_name()

            # create and add new Rms Events Group
            new_group = RmsEventsGroup(idtag=None, name=name)
            self.circuit.add_rms_events_group(new_group)

            # update event rows
            for row in self.rows:
                row.add_group(new_group)

            QtWidgets.QMessageBox.information(
                self,
                "Group Created",
                f"New group name: {name}"
            )

    def accept_dialog(self):
        """Validate and collect event data."""

        if not self.rows:
            QtWidgets.QMessageBox.warning(
                self,
                "No Events",
                "Please add at least one event before confirming.",
            )
            return

        parameters = []
        target_times = []
        values = []
        groups = []

        for i, row in enumerate(self.rows):

            try:
                param, t, v, group = row.get_data()

            except Exception as exc:

                QtWidgets.QMessageBox.warning(
                    self,
                    "Invalid Input",
                    f"Row {i + 1}: {exc}",
                )
                return

            parameters.append(param)
            target_times.append(t)
            values.append(v)
            groups.append(group)
        self.data["parameters"] = parameters
        self.data["target_times"] = target_times
        self.data["values"] = values
        self.data["groups"] = groups
        self.accept()

    def get_data(self) -> Dict[str, List]:
        """Return collected data."""
        return self.data

class RmsEventsGroupsDialog(QtWidgets.QDialog):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle("Create RMS Events Group")
        self.setModal(True)
        self.setMinimumWidth(300)

        self._name: str = None

        # Widgets
        self.name_label = QtWidgets.QLabel("Name:")
        self.name_edit = QtWidgets.QLineEdit()
        self.name_edit.setPlaceholderText("Enter group name")

        # Buttons
        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel
        )

        # Layout
        form_layout = QtWidgets.QFormLayout()
        form_layout.addRow(self.name_label, self.name_edit)

        main_layout = QtWidgets.QVBoxLayout(self)
        main_layout.addLayout(form_layout)
        main_layout.addWidget(self.buttons)

        # Connections
        self.buttons.accepted.connect(self.accept_dialog)
        self.buttons.rejected.connect(self.reject)
        self.name_edit.returnPressed.connect(self.accept_dialog)

    def accept_dialog(self):
        """
        Validate the input and store the name
        """
        name = self.name_edit.text().strip()

        if not name:
            QtWidgets.QMessageBox.warning(
                self,
                "Invalid name",
                "The name cannot be empty."
            )
            return

        self._name = name
        self.accept()

    def get_name(self) -> str:
        """
        Returns the entered name
        """
        return self._name




